SystemOrganization addCategory: #'CommandShell-Base'!

Model subclass: #CommandShell
	instanceVariableNames: 'outputStream errorStream shellSyntax environment builtinCommands commandHistory historyIndex promptString promptStringTwo foregroundProxies backgroundProxies backgroundPipelines commandLineProcesses backgroundMessageQueue keyboardConnector activeController safeToExit readSyncSemaphore'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Base'!

!CommandShell commentStamp: '<historical>' prior: 0!
I am a command shell, similar to /bin/sh, with a simple command line user interface. I collaborate with process proxies to provide command execution, and I provide a limited set of built in commands similar to those in /bin/sh. My built in commands are implemented in Smalltalk, and any other commands are passed to process proxies to be executed either internally as Smalltalk "doIt" expressions, or externally as commands passed to the external operating system. I am similar to a TranscriptStream (some methods are copied directly from TranscriptStream), but I also know how to accept lines of command input, parse them, and hand them off to process proxies for execution.

Three types of commands may be executed from a CommandShell: internal "builtin" commands implemented in Smalltalk; internal Smalltalk "doIt" commands; and external commands. Internal commands (builtin commands or doIt commands) may be freely mixed with external operating system commands in a command pipeline. See CommandShell class>>commandProcessing for more information.

Each command line is first evaluated as a Smalltalk expression, and is subject to further parsing only if the Smalltalk evaluation fails. In practice, this permits complete Smalltalk expressions to be evaluated easily without conflicting with shell syntax, and allows Smalltalk and unix shell commands to be freely mixed.

Simple command scripting is supported (method category 'command scripting'). Any mix of internal and external commands may be included in a script. Conditional branching is supported based on command exit status.

Open a new shell window with "CommandShell open". Type 'help' followed by <return> or <enter> for help on builtin commands.

Things that work reasonably well:
- Simple command execution for running command line programs or starting
  X programs.
- Command pipelines. Built in commands can be mixed with external
  commands, as in "help sqsh | wc -l".
- Command IO redirection with '<',  '>', '>>', '2>', and '2>>'.
- Command history and command history recall.
- Background command execution, as in "xterm&".
- <ctl-C> to interrupt a running external command.
- <ctl-D> to indicate end of file in terminal input.

Limitations include:
- Dumb tty only. Do not try to run vi.
- Standard Unix shell syntax is not completely implemented.

Race conditions are possible for certain command pipelines. See CommandShell class>>raceConditions for more information.
!

!CommandShell class methodsFor: 'instance creation' stamp: 'dtl 12/11/2007 21:22'!
command: aCommandString
	"Evaluate aCommandString in the context of a CommandShell, using a
	CommandShellTranscript to display the results."

	"CommandShell command: 'help'"
	"CommandShell command: 'ls'"

	| cs w |
	cs := self new.
	w := cs openLabel: self defaultWindowName.
	w scheduleToEvaluate: [w nextPutAll: aCommandString; cr].
	w processCommand: aCommandString.
	^ w
! !

!CommandShell class methodsFor: 'documentation' stamp: 'dtl 12/29/2001 14:18'!
commandProcessing
	"Explanation of the command parsing and evaluation."

	"CommandShell new pipeline: 'CommandShell commandProcessing!! | edit'"

	^
'CommandShell behaves similarly to a Unix command shell. One or more Morphic or MVC views can be opened on an instance of CommandShell. Command lines are accepted in the view, passed to the CommandShell instance (the model), and the results are displayed in the view(s). The overall behavior can be thought of as a Transcript acting like an xterm window, and the CommandShell acting like a /bin/sh shell.

A command line is first evaluated as a Smalltalk expression if possible. If this evaluation fails, either as a result of a sytax error or a runtime error, it is assumed to be a command string with syntax similar to that of a Unix command shell.

The command line is parsed by CommandShell to break command lines into elements of a command pipeline. All further parsing, such as file name expansion, is performed by an instance of ShellSyntax.

Three types of commands may be executed from a CommandShell:
1) Internal "builtin" commands are implemented in Smalltalk, and behave similarly to the shell builtin commands of a typical Unix command shell. See protocol category "shell builtins" for the implementation of these builtin shell commands.
2) Internal "doIt" commands are simple Smalltalk expression strings, terminated by a $!! character, which are evaluated as Smalltalk expressions, and which recognize the tokens "stdin", "stdout", and "stderr" as temporary variables in the doIt expression. An internal doIt command is evaluated such that it may be incorporated into a command pipeline with access to the stdin/stdout/stderr streams used in the command pipeline. In the current implementation, a doIt command may not include a $| character, which has meaning both for Smalltalk syntax and for Unix command line processor syntax.
3) External commands. Any command expression not recognized as a shell builtin or as a doIt expression is passed to an external process proxy for execution by the external operating system, with stdin/stdout/stderr incorporated into the command pipeline.

Internal and external commands can be freely mixed in a command pipeline, with results which should be familiar to users of Unix command shells. Pipes are implemented as instances of OSPipe or InternalPipe, and are capable of passing streams of characters from one process proxy to the next. All commands (internal or external) evaluate to strings, such that their results can either be passed through a pipe to the next command proxy in a pipeline, or displayed in the terminal window view if the proxy represents the last command in a command pipeline. Error messages (also strings) are written to an errorPipelineStream which is shared by all proxies in a command pipeline, such that any error messages are accumulated for the entire command pipeline.'! !

!CommandShell class methodsFor: 'defaults' stamp: 'dtl 4/2/2001 20:48'!
defaultWindowName

	^ 'Squeak Shell'! !

!CommandShell class methodsFor: 'documentation' stamp: 'dtl 12/15/2007 12:32'!
htmlOverview
	"A general overview"

	^
'<html><head><title>Introduction to CommandShell</title></head>
<body>
<h3>CommandShell</h3>
 CommandShell is a Smalltalk implementation of a command processor shell and terminal
 window. It is intended to behave like to a simple terminal window (like xterm) running
 a Unix command shell (like /bin/sh). It lacks some elements of Unix shell syntax, and
 does not provide terminal emulation, but it adds some nice Smalltalk enhancements such
 as a text editor which works in a command pipeline, and the ability to evalute Smalltalk
 expressions in a command pipeline with Unix commands.

<h3>How to use CommandShell</h3>
  Evaluate "CommandShell open" or, if you have loaded the change set to put CommandShell
  in your world menu, just select "squeak shell" from the "open..." menu entry.
  To use the command shell window, enter commands on the command line (after
  the "$ " command prompt, followed by cr to accept the command. The "help" command
  gives some clues as to what can be done. In general, just try any command which
  you might otherwise enter into a Unix shell or xterm window. Then try some of
  the Smalltalk features such as mixing Smalltalk expressions with Unix commands, and
  piping commands into Smalltalk editors.

<h3>Basic Concepts</h3>
<ul><li><h4>  Commands</h4>
    Smalltalk uses objects and messages, with expressions arranged in a natural
    Noun-Verb-Predicate sentence structure, similar to that of some spoken languages.
    In contrast, many other computer systems use a style in which users issue
    a command to the system, in which the "command" (a verb) is followed by various
    (predicate) modifiers. In the case of a traditional command-line interface, such
    as a Unix shell, the command is usualy the name of a program, and the modifiers
    are parameters passed to the program to influence its behavior. CommandShell attempts
    to provide a command line interface for executing commands within Squeak, in
    a style and environment which is comforable for someone accustomed to the
    Smalltalk style of expression.
<p>
<li><h4>  Process Proxies</h4>
    A command line user interface, such as a Unix shell or a "DOS window" in
    Windows, revolves around the notion of executing programs from user commands.
    When Squeak is hosted in another operating system, it is useful to be able to
    execute an external program from a command line expression. A process proxy
    is an object which represents the evaluation of such an external command
    in a command line environment.
<p>
    From the point of view of Squeak, a process proxy might just as well represent
    the evaluation of a Smalltalk expression in the context of a command line
    environment. Thus a process proxy can be an object which represents the execution
    of an external program or an internal Smalltalk expression. It is created from a
    command line expression, and provides a representation of certain aspects of the
    external (or internal) process execution, such as process run state, exit status, and
    the input, output and error streams for the process.
<p>
    External process proxies are used to evaluate programs in the external operating
    system (this requires the OSProcess change set in addition to CommandShell). Internal
    process proxies are used to evaluate "built in" commands for a command shell,
    such as the "cd" command to change the working directory for the command shell.
    Internal proxies are also used for evaluating "doIt" expressions within a command
    pipeline. In this case, the variables "stdin", "stdout", and "stderr" are pre-defined
    such that the doIt expression can directly access the input, output and error
    pipes for the internal process proxy (in other words, the doIt expression
    "stdout nextPutAll: ''Hello world'' !!" will write ''Hello world'' on the output pipe
    of the proxy).
<p>
<li><h4> Pipelines</h4>
    Powerful computing systems may be built upon simple design metaphors. In
    Smalltalk, the concept of objects communicating through messages is generalized
    and extended to produce the Squeak system. In Unix systems, the metaphor of a
    pipe with data flowing between two programs is generalized to produce a mechanism
    for connecting small programs to produce complex systems. The command line shells
    for Unix support this metaphor by providing syntax for connecting two or more
    commands together into more complex command pipelines.
<p>
    CommandShell provides a framework for connecting several process proxies into a
    command pipeline, using command syntax similar to that of a Unix shell. Since the
    process proxies are implemented in Smalltalk, the "programs" in a command pipeline
    may consist of process proxies representing external programs, or of proxies representing
    the evaluation of internal Smalltalk expressions. The objects which flow through
    the pipes in a command pipeline are assumed to be characters, such that the external
    commands can operate on the streams of characters in the usual way, and the internal
    Smalltalk commands read and write streams of characters.
<p>
    Just as programs (or internal expressions) are represented by process proxies, the
    pipes which interconnect two process proxies are represented by pipe objects. The
    pipe may be either an OSPipe (a proxy representing a Unix system FIFO pipe) or an
    InternalPipe (a Smalltalk object which behaves similarly to an OSPipe). The combination
    of process proxies and pipes is a command pipeline, which may be created and
    evaluated from a command line expression using conventional Unix shell syntax.
</ul>
<h3>Command Line Syntax</h4>
    A command or command pipeline may be created from a string expression, and
    evaluated using Unix shell syntax. CommandShell accepts command line strings and
    does some high level processing to set up process proxy execution. For most of the
    command parsing, it relies on a ShellSyntax object to do the syntax evaluation.
    In particular, a real Unix shell interprets command line parameters in the context of the
    Unix file system, expanding "wildcard" characters and searching for files in the context
    of a "current working directory" location in the file system tree. A CommandShell,
    collaborating with an instance of ShellSyntax, implements this syntax parsing
    and evaluation in Smalltalk.

<h3>Command Execution</h4>
    Beginning with a command line (which may represent a pipeline of several
    commands), a CommandShell first attempts to treat the entire command line as
    a Smalltalk expression. The command line string is used to create an instance of
    PipeableEvaluator (a kind of process process). If the expression is successfully
    compiled and evaluated in the PipeableEvaluator, the command line is executed
    just as if it were an external Unix command, with the result of the evaluation
    written as a string to the output stream of the process proxy.
<p>
    If the complete command line cannot be treated as a Smalltalk expression,
    CommandShell assumes that it must be a Unix style command pipeline. If the
    command line is a comment (starting with ''#'') it is discarded; otherwise it
    is broken down into a series of process proxies connected by pipes, with
    command line parameters and other aspects of the command line parsing handled
    by a ShellSyntax object.
<p>
    The process proxies in the command pipeline may represent internal or external
    commands. CommandShell parses the command line to create the proxies by breaking
    the command pipeline into segments (separated by the "|" pipe character), with
    different kinds of process proxy created according to the following priorities:
<ul><li>
      If the command segment is a simple Smalltalk expression terminated by "!!", it
      is treated as a "doIt" expression, evaluated in a PipeableEvaluator proxy. This
      kind of simple doIt expression is limited to commands with characters which
      are unambiguous in a shell command. For example, "|" has special meaning in
      a command line, and cannot be used in a doIt expression (command line quoting
      and escape characters have not yet been implemented in the CommandShell syntax).
<li>
      If the command segment is not a doIt expression, CommandShell checks to see
      if it matches one of several "shell builtin" commands. These are special commands
      implemented in class ShellBuiltin and evaluated in a PipeableEvaluator proxy.
      They behave analogously to shell builtin commands in a conventional Unix shell.
      A number of commands such as "cd" and "pwd" are implemented as internal builtin
      commands, rather than relying on Unix external equivalents. In addition, other
      builtin commands such as "edit" do things which are useful in Squeak and have
      no real equivalent in Unix (the "edit" builtin opens an editor within Squeak, taking
      its input from files or directly from the output of a command pipeline such as
      "Smalltalk !! | edit" or "who | edit").
<li>
      If the command segment is not a shell builtin command, it is assumed to be an
      external command. It this case, the command line segment is fully parsed, then
      used to create an external OS process proxy to run the external command. The
      external process proxy responds to a #value message by running the external
      program, with output and error written to pipes connected to the command
      pipeline.
<li>
      Finally, if no external program can be found to execute the command, an error
      message is provided to the command window, and no command is evaluated.
</ul>
    Once the command proxy pipeline has been created, the individual process proxies
    are evaluated in such a way that they appear to execute in parallel, with each
    proxy reading its input from a pipe connected to its predecessor in the pipeline,
    and writing its output to a pipe connected to its successor in the pipeline. An
    additional error pipe is shared by all process proxies in the pipeline, such that
    any error output is accumulated in the shared error pipe stream.
<p>
    A command shell can obtain the output and error of a complex command pipeline
    simply by evaluating the last process proxy in the chain, and reading up to the
    end of its output and error pipes. Since a process proxy knows its run state, and
    a pipe is not "at end" until a process proxy closes one end of the pipe, the pipeline
    can be constructed in such a way that the command shell is assured that all process
    proxies have completed their evaluation when the output of the last proxy is read
    up to the end of the pipe.
<p>
    As the last process proxy in a command pipeline is evaluated, CommandShell reads
    its output and error pipes, and displays the text in the View or Morph which it uses
    as its terminal window. When complete, it issues a new prompt string, and waits for
    another command line to be entered by the user.
<p>
<h3>Exercise for the Reader:</h4>
Which one of the following five command lines will generate
an error, and why?
<ol><li>
  $ stdout nextPutAll: ''hello world''
<li>
  $ stdout nextPutAll: ''hello world''; cr
<li>
  $ stdout nextPutAll: ''hello world''!!
<li>
  $ stdout nextPutAll: ''hello world''; cr!!
<li>
  $ stdout nextPutAll: ''hello world''!!; stdout cr!!
</ol>
<h3>Answer:</h4>
The fourth command will generate an error.
<ul><li>
  Line one is evaluated as a complete Smalltalk expression, and succeeds.
<li>
  Line two is evaluated as a complete Smalltalk expression, and succeeds.
<li>
  Line three is evaluated as a pipeline after failing evaluation as a complete Smalltalk
  expression. It succeeds as a pipeline with one internal doIt proxy, and produces
  the same output as line one.
<li>
  Line four cannot be evaluated as a complete Smalltalk expression. The ";" token
  causes the command line to be evaluated as two pipelines. The expression for the
  first pipeline does not have a trailing "!!" token, so it is assumed to be an external
  command, and fails. The second pipeline is evaluated as a doIt on the string ''cr!!'',
  which fails as an invalid Smalltalk expression. The error messages for both failures
  are accumulated on the shared error pipe, and are displayed in the terminal window.
<li>
  Line five is evaluated successfully as two pipelines, each consisting of one doIt
  proxy. The output is the same as for line two.
</ul></body></html>
'! !

!CommandShell class methodsFor: 'class initialization' stamp: 'dtl 9/3/2010 17:05'!
initialize

	"CommandShell initialize"

	| preferencesClass |
	Smalltalk addToStartUpList: self after: ImageSegment.
	(preferencesClass := Smalltalk at: #Preferences)
		ifNotNil: ["Pharo deprecates Preference in favor of some other mechanism"
			(preferencesClass respondsTo: #setWindowColorFor:to: )
				ifTrue: ["Squeak"
					preferencesClass 
						setWindowColorFor: #CommandShell 
						to: (Color colorFrom: self windowColorSpecification pastelColor)]
				ifFalse: ["Pharo images use some other preferences protocol"]].
	(TheWorldMenu respondsTo: #registerOpenCommand:)
		ifTrue: [TheWorldMenu registerOpenCommand: {'Squeak shell'. {CommandShell . #open}}]
		ifFalse: ["Pharo menu registration uses pragmas, would break on Squeak 3.8, do not use it here"]
! !

!CommandShell class methodsFor: 'version testing' stamp: 'dtl 6/29/2010 21:06'!
isMorphic
	"Squeak traditionally used #isMorphic to distinguish between Morphic
	and MVC user interfaces. In newer Squeak images, the current project
	knows its active user interface, so requests are vectored through the
	current project. Pharo images pretend to be unaware of the fact that
	they are running Morphic, refusing to respond to #isMorphic. This 
	implementation does reasonable things in any of the above scenarios."

	^(Smalltalk respondsTo: #isMorphic) not or: [Smalltalk isMorphic]! !

!CommandShell class methodsFor: 'VM building' stamp: 'dtl 8/4/2002 17:15'!
makeVmIn: buildDirectoryPathName
	"Rebuild the virtual machine and plugins in the buildDirectoryPathName
	directory. If the build is successful, save the image and restart using the
	new VM. This assumes that the currently executing VM is either located in,
	or linked to, the buildDirectoryPathName directory."

	"CommandShell makeVmIn: FileDirectory default pathName, FileDirectory slash, 'build'"

	| shell |
	shell := self new.
	shell open.
	^ shell makeVmIn: buildDirectoryPathName
! !

!CommandShell class methodsFor: 'instance creation' stamp: 'dtl 12/27/2001 23:36'!
open
	"CommandShell open"

	^ self new open! !

!CommandShell class methodsFor: 'instance creation' stamp: 'dtl 3/17/2003 21:19'!
openLabel: aString

	"CommandShell openLabel: self defaultWindowName"

	^ self new openLabel: aString
! !

!CommandShell class methodsFor: 'instance creation' stamp: 'dtl 6/27/2010 16:08'!
openWithHelp
	"CommandShell openWithHelp"

	self open processCommand: 'help'
! !

!CommandShell class methodsFor: 'documentation' stamp: 'dtl 12/1/2001 18:51'!
overview
	"A general overview"

	"CommandShell new pipeline: 'CommandShell overview!! | edit'"

	| rs ws |
	rs := ReadStream on: self htmlOverview.
	ws := WriteStream on: String new.
	[rs atEnd]
		whileFalse:
			[ws nextPutAll: (rs upTo: $<).
			rs upTo: $>].
	^ ws contents
! !

!CommandShell class methodsFor: 'proxy creation' stamp: 'dtl 12/31/2007 14:51'!
pipeableProxyFor: commandString 
	"Create a new proxy for a PipeJunction. The proxy may be any type
	of PipeJunction, such as a PipeableOSProcess or ExpressionEvaluator.
	Send #value to evaluate the newly created proxy."

	"(CommandShell pipeableProxyFor: 'ls -l') value upToEndOfFile"
	"(CommandShell pipeableProxyFor: 'Array new: 4 !!') value upToEndOfFile"
	"(CommandShell pipeableProxyFor: 'help') value upToEndOfFile"

	^ self new
		pipeableProxyFor: commandString
		input: nil
		output: nil
		error: nil
		predecessorProxy: nil! !

!CommandShell class methodsFor: 'proxy creation' stamp: 'dtl 12/31/2007 14:55'!
pipeline: aCommand
	"Find the individual commands in a command pipeline string, and evaluate
	them as a command pipeline. Answer a ProxyPipeline collection of processes
	in the pipeline."

	"(CommandShell pipeline: 'ps | grep ps | wc') upToEndOfFile"
	"(CommandShell pipeline: 'ls NOSUCHFILE * | dd | cat') errorUpToEndOfFile"
	"(CommandShell pipeline: 'ls NOSUCHFILE * | dd | cat') upToEndOfFile"

	^ self new pipeline: aCommand
! !

!CommandShell class methodsFor: 'documentation' stamp: 'dtl 12/15/2007 12:25'!
raceConditions

	"CommandShell new pipeline: 'CommandShell raceConditions!! | edit'"

	^
'It may be possible to hang the Squeak VM in an IO race condition when doing certain command pipelines. These conditions have been largely eliminated for Unix platforms through the use of nonblocking pipes in conjunction with asynchronous IO or polling processes on the Smalltalk side of the pipes. However, these notes are provided to describe possible deadlock scenarios.

Note that nonblocking OS pipes are not yet available on Windows. Command pipelines with external OS processes on Windows are not supported. Blocking on read or write to an OSPipe in Windows will definitely lock the VM.

Two kinds of deadlock are possible. In both cases, deadlock occurs when the single-threaded Squeak VM is blocked on a read or write to an OS pipe, and the blocked condition can only be cleared by another process running in the blocked VM.

1) Block on write. The Squeak VM can block on a write to an OS pipe if the external process at the other end of the pipe does not keep up with reading it. One internal process proxy writes on an pipe to an external process proxy, and fills the pipe up to some capacity limit, after which the write blocks. The external proxy is writing its output to another internal process proxy in the command pipeline, which never executes because the VM is still blocked trying to write to the full pipe.

2) Block on read. The Squeak VM can block on a read from an OS pipe if the external process writing to the pipe never closes the pipe. The internal proxy will post a read on the output pipe from the external command. If the external command has exited, the read will return with a (possibly empty) string. However, if the external command either does not exit, or exits without Squeak being notified, the read on the output pipe will block indefinitely.

2a) In the case where the external command exits while the Squeak VM is blocked on a read, Squeak is unable to respond to the notification of the external process exiting. The process stays in a zombie state, the pipes therefore never get closed, and the VM stays in a blocked condition.

Here are things to try if a deadlock occurs:

Squeak will be blocked and will not respond to the mouse or keyboard. Open a command shell outside of Squeak (perhaps an xterm, or log in to another virtual terminal on Linux or FreeBSD). Using the "ps" command, find the Squeak VM process and its subprocesses. Kill the subprocess. In many cases this will free up the deadlock, and Squeak will come back to life. If this does not work, then you will probably find that one of the processes which you just killed is now in a "zombie" state, which means that the Squeak VM is deadlocked in such a way that it cannot recognize that its child process exited (and therefore cannot properly close the associated file handles). In this case, if your operating system supports a /proc file system, you may be able to break things free by emptying the pipes. On Linux, go to /proc/<pidOfTheSqueakVMProcess>/fd/ and do "cat" commands on all the pipe files in this directory. If none of this works, you are just plain stuck, and you will need to kill your Squeak session and recover anything valuable from the changes log.'! !

!CommandShell class methodsFor: 'system startup' stamp: 'dtl 7/12/2002 19:26'!
startUp: resuming
	"A CommandShell may be running external processes in background, and waiting
	for those processes to complete. If the Squeak image is stopped and restarted, the
	background processes are no longer children of the Squeak process, and there is
	no way to receive notification when the processes exit. This would result in leftover
	Smalltalk processes associated with the CommandShell waiting forever for the
	external processes to exit. The simplest way to prevent this is to terminate all
	the Smalltalk processes unconditionally after an image restart."

	resuming ifTrue: [self allSubInstances do: [:e | e startUpInNewSession]]
! !

!CommandShell class methodsFor: 'version testing' stamp: 'dtl 9/3/2010 17:48'!
versionString

	"CommandShell versionString"

	^'4.4.9'! !

!CommandShell class methodsFor: 'window color' stamp: 'dtl 6/29/2010 00:07'!
windowColorSpecification
	"Answer a WindowColorSpec object that declares my preference"

	| windowColorSpec |
	windowColorSpec := Smalltalk
				at: #WindowColorSpec
				ifAbsent: [^ self error: 'this image does not support WindowColorSpec'].
	^ windowColorSpec
		classSymbol: self name
		wording: 'Squeak Shell'
		brightColor: (Color lightGray lighter paler)
		pastelColor: (Color lightGray lighter lighter paler paler)
		helpMessage: 'CommandShell window for evaluating Smalltalk and OS commands'! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/24/2001 19:09'!
activeController
	"In Morphic, alway nil. In MVC, the controller which most recently invoked
	a command."

	^ activeController! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/24/2001 19:09'!
activeController: aController
	"In Morphic, alway nil. In MVC, the controller which most recently invoked
	a command."

	activeController := aController! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/19/2002 07:13'!
backgroundMessageQueue
	"When a background pipeline completes, it places a message in this queue.
	The message will be displayed the next time the user accepts a line of
	text (presses CR)."

	^ backgroundMessageQueue ifNil: [backgroundMessageQueue := OrderedCollection new]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 8/19/2006 13:14'!
backgroundMessageQueue: anObject
	"Set the value of backgroundMessageQueue"

	backgroundMessageQueue := anObject! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 12/11/2007 21:33'!
backgroundPipelineComplete: aPipeline commandIndex: index
	"This message is sent when aPipeline has completed evaluation.
	Ensure that the output and error event handlers have had a chance
	to completely empty the pipes prior to cleaning things up."

	self waitUntilPipesAreEmpty: aPipeline.
	aPipeline closePipes; release.
	self backgroundMessageQueue addLast:
		((WriteStream on: String new)
			nextPutAll: '[', index asString, ']+ Done                     ';
			nextPutAll: aPipeline commandLine;
			cr;
			contents).
	self backgroundProxies removeAll: aPipeline.
	self unRegisterBackgroundPipeline: aPipeline at: index.
	self notifyRestorePrompt
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/15/2002 16:23'!
backgroundPipelines
	"The pipelines which are currently being evaluated in background. This
	collection is used to keep track of the background command lines by giving
	each background pipeline an index number corresponding to its position in
	this collection. New pipelines are inserted at the first available nil position
	in the collection, and the collection is expanded as needed."

	^ backgroundPipelines ifNil: [backgroundPipelines := OrderedCollection new]

! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 8/19/2006 13:14'!
backgroundPipelines: anObject
	"Set the value of backgroundPipelines"

	backgroundPipelines := anObject! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/9/2006 10:23'!
backgroundProxies
	"All of my process proxies that are being evaluated in the background."

	backgroundProxies ifNil: [backgroundProxies := OrderedCollection new].
	^ backgroundProxies! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/9/2006 10:23'!
backgroundProxies: aCollection
	"All of my process proxies that are being evaluated in the background."

	backgroundProxies := aCollection
! !

!CommandShell methodsFor: 'process proxy creation' stamp: 'dtl 11/21/2006 09:20'!
builtinCommandProxyFor: aCommandString input: in output: out error: err
	"Answer a PipeableEvaluator on a built in command, or nil."

	"CommandShell new initialize builtinCommandProxyFor: 'help' input: nil output: nil error: nil"

	| args command commandAndArgs |
	^ (self isBuiltInCommand: aCommandString)
		ifTrue:
			[commandAndArgs := self shellSyntax
									programNameAndArgumentsFrom: aCommandString
									inDirectoryPath: nil
									findExecutable: false.
			command := self builtinCommands at: commandAndArgs first.
			args := commandAndArgs last.
			^ ShellBuiltinEvaluator
				command: command
				withArguments: args
				shell: self
				pipeFrom: in
				pipeTo: out
				errorPipelineStream: err]
		ifFalse: [nil]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/27/2001 12:24'!
builtinCommands

	^ builtinCommands ifNil: [builtinCommands := Dictionary new]! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 7/14/2002 11:54'!
builtinCommands: aCommandDictionary

	builtinCommands := aCommandDictionary! !

!CommandShell methodsFor: 'testing' stamp: 'dtl 8/7/2002 11:45'!
canAccessOSProcess
	"True if OSProcess is in the image and if the OSProcess plugin is functioning properly"

	"CommandShell new canAccessOSProcess"

	| osp |
	^ (osp := Smalltalk at: #OSProcess ifAbsent: []) notNil and: [osp accessor canAccessSystem]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 3/16/2001 05:58'!
characterLimit
	"Tell the views how much to retain on screen"
	^ 20000! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 2/18/2007 15:37'!
checkSttyFor: characterValue
	"Check for interrupt characters and such. Consume interrupt character and
	answer nil, otherwise answer characterValue."

	(self isInterrupt: characterValue)
		ifTrue:
			[self changed: #interruptCharacter.
			^ nil].
	(self isEndOfFile: characterValue)
		ifTrue:
			[self doEndOfFile.
			^ nil].
	^ characterValue! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/2/2001 12:05'!
clearSemaphore

	[self readSyncSemaphore isEmpty] whileFalse: [readSyncSemaphore terminateProcess].
	readSyncSemaphore signal
! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 1/13/2007 18:04'!
command: aCommandString
	"Evaluate aCommandString. Answer a collection of evaluated pipelines,
	or nil if nothing was done."

	"CommandShell new command: 'ls -l | cat'"
	"CommandShell new open; command: 'ls -l | cat'"
	"CommandShell new open; command: 'who'"
	"CommandShell new open; command: 'who&'"
	"CommandShell new open; command: 'sol'"
	"CommandShell new command: 'sol'"

	^ self evaluatePipelines: (self evaluateOrMakePipelinesFrom: aCommandString)
! !

!CommandShell methodsFor: 'command scripting' stamp: 'dtl 3/18/2003 20:33'!
command: aCommandString onFailureDo: aBlockWithZeroOrOneParameter
	"Evaluate aCommandString. On failure, answer the result of evaluating
	aBlockWithOneParameter with the collection of evaluated pipelines as
	its parameter. Otherwise, answer a collection of evaluated pipelines or
	nil if nothing was done."

	"CommandShell new open;
		command: 'who'
		onFailureDo: [:p | 'the proxy pipeline collection which failed was ', p printString]"
	"CommandShell new open;
		command: 'ls NOSUCHFILE'
		onFailureDo: [:p | 'the proxy pipeline collection that failed was ', p printString]"
	"CommandShell new open;
		command: 'ls NOSUCHFILE'
		onFailureDo: ['the command failed']"

	| pipelines |
	pipelines := self command: aCommandString.
	(pipelines anySatisfy: [:p | p succeeded not])
		ifTrue:
			[(aBlockWithZeroOrOneParameter numArgs == 0)
				ifTrue: [^ aBlockWithZeroOrOneParameter value]
				ifFalse: [^ aBlockWithZeroOrOneParameter value: pipelines]]
		ifFalse: [^ pipelines]

! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 4/7/2001 12:48'!
commandHistory
	"Automatically trim the history list, and answer the resulting list."

	| trimIncrement |
	trimIncrement := 1.	"Could be a larger number to copy the collection less often"
	commandHistory isNil
		ifTrue:
			[commandHistory := OrderedCollection new]
		ifFalse:
			[(commandHistory size > (self historyDepth * 2 + (trimIncrement * 2) - 2))
				ifTrue:
					[commandHistory := commandHistory
						copyFrom: (trimIncrement * 2 + 1) to: commandHistory size]].
	^ commandHistory! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/15/2002 09:21'!
commandLineProcesses
	"The Smalltalk processes associated with currently executing command lines. Keep
	track of them here so that they can be conveniently terminated if something gets
	out of whack and needs cleaning up."

	commandLineProcesses ifNil: [commandLineProcesses := OrderedCollection new].
	^ commandLineProcesses
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 8/19/2006 13:14'!
commandLineProcesses: anObject
	"Set the value of commandLineProcesses"

	commandLineProcesses := anObject! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 1/13/2007 19:10'!
copyAllErrorFrom: aProxy
	"An event driven data source may send this message to cause its available
	error stream characters to be processed."

	| s |
	self readSyncSemaphore critical:
		[s := aProxy errorUpToEnd.
		s isEmpty ifFalse:
			[[self errorStream nextPutAll: s]
				on: Error
				do: [:ex | ex retryUsing:
						["If the error pipe overflows we get an unhandled error,
						so wait a while and retry the write."
						(Delay forMilliseconds: self pollingDelayTime) wait.
						self errorStream nextPutAll: s]]]]
! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 1/16/2007 07:05'!
copyAllOutputFrom: aProxy
	"An event driven data source may send this message to cause its available
	characters to be processed."

	| s |
	self readSyncSemaphore critical:
		[s := aProxy upToEnd.
		s isEmpty ifFalse:
			[[self outputStream nextPutAll: s]
				on: Error
				do: [:ex | ex retryUsing:
						["If the output pipe overflows we get an unhandled error,
						so wait a while and retry the write."
						(Delay forMilliseconds: self pollingDelayTime) wait.
						self outputStream nextPutAll: s]]]]
! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 1/13/2007 19:10'!
copyErrorToEofFrom: aProxy
	"An event driven data source may send this message to cause its available
	error stream characters to be processed."

	| s |
	self readSyncSemaphore critical:
		[s := aProxy errorUpToEndOfFile.
		s isEmpty ifFalse:
			[[self errorStream nextPutAll: s]
				on: Error
				do: [:ex | ex retryUsing:
						["If the error pipe overflows we get an unhandled error,
						so wait a while and retry the write."
						(Delay forMilliseconds: self pollingDelayTime) wait.
						self errorStream nextPutAll: s]]]]
! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 1/13/2007 19:11'!
copyOutputToEofFrom: aProxy
	"An event driven data source may send this message to cause its available
	characters to be processed."

	| s |
	self readSyncSemaphore critical:
		[s := aProxy upToEndOfFile.
		s isEmpty ifFalse:
			[[self outputStream nextPutAll: s]
				on: Error
				do: [:ex | ex retryUsing:
						["If the output pipe overflows we get an unhandled error,
						so wait a while and retry the write."
						(Delay forMilliseconds: self pollingDelayTime) wait.
						self outputStream nextPutAll: s]]]]
! !

!CommandShell methodsFor: 'defaults' stamp: 'dtl 7/14/2002 12:42'!
defaultBuiltinCommands

	^ #(cd: clear: copyToError: copyToOutput: edit: snapshot: exit: fc: help: history: inspect: pwd: sls: sqsh: type:)
! !

!CommandShell methodsFor: 'defaults' stamp: 'dtl 6/17/2002 23:43'!
defaultPromptString

	^ '$ '! !

!CommandShell methodsFor: 'defaults' stamp: 'dtl 6/17/2002 23:45'!
defaultPromptStringTwo

	^ '> '! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 9/3/2010 17:02'!
doEndOfFile
	"EOF character detected, close stdin to pipeline."

	| pipelineInput |
	(pipelineInput := self keyboardConnector)
		ifNotNil: [pipelineInput closeWriter]
! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 3/9/2002 18:22'!
doInterrupt
	"Interrupt character detected, do interrupt stuff."

	| msgStrm children answer |
	children := self externalChildren collect: [:e | e processProxy].
	(children size > 0)
		ifTrue:
			[msgStrm := WriteStream on: String new.
			(children size > 1)
				ifTrue: [msgStrm nextPutAll: 'kill processes']
				ifFalse: [msgStrm nextPutAll: 'kill process'].
			children do: [:e | msgStrm nextPutAll: ' ', e pid printString, ' (', e programName, ')'].
			msgStrm nextPut: $?.
			self readSyncSemaphore critical:
				[self pauseStProcesses.
				answer := self confirm: msgStrm contents.
				self unpauseStProcesses.
				answer ifTrue:
					[children reverseDo: [:e | e sigkill].
					self terminateStProcesses]]]
! !

!CommandShell methodsFor: 'process proxy creation' stamp: 'dtl 1/20/2007 14:18'!
doItProxyFor: aCommandString input: in output: out error: err
	"Answer a PipeableEvaluator on a doIt command, or nil."

	"CommandShell new doItProxyFor: '#thisIsADoIt !! ' input: nil output: nil error: nil"

	^ (self shellSyntax isDoItCommand: aCommandString)
		ifTrue: [self evaluationProxyFor: (self shellSyntax doItStringFrom: aCommandString)
					input: in
					output: out
					error: err]
! !

!CommandShell methodsFor: 'command history' stamp: 'dtl 7/1/2001 18:25'!
editCommand: aString

	^ FillInTheBlank
		request: 'edit command'
		initialAnswer: aString! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 6/7/2001 23:17'!
endOfFileCharacterValue
	"<ctl>-D"

	^ 4! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 5/18/2008 11:36'!
environment

	^ environment ifNil:
		[environment := (Smalltalk at: #OSProcess ifPresent:
			[:osp | osp thisOSProcess environment deepCopy]) ifNil: [Dictionary new]].
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 8/19/2006 13:14'!
environment: anObject
	"Set the value of environment"

	environment := anObject! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 1/16/2007 08:08'!
errorStream
	"Answer the value of errorStream"

	^ errorStream ifNil: [errorStream := WriteStream on: '']! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 8/19/2006 13:14'!
errorStream: anObject
	"Set the value of errorStream"

	errorStream := anObject! !

!CommandShell methodsFor: 'command line parsing' stamp: 'dtl 1/22/2007 07:04'!
evaluateOrMakePipelinesFrom: aCommandString
	"Convert aCommandString into a structured collection of process proxies,
	possibly after evaluating aCommandString as Smalltalk."

	"CommandShell new evaluateOrMakePipelinesFrom: 'who | cat | wc; ps -aef | cat& pwd | cat'"
	"CommandShell new evaluateOrMakePipelinesFrom: 'Array new: 10' "

	| process |
	"Ignore trivial case of a single word starting with $#"
	(self isOneWordCommentString: aCommandString)
		ifTrue:
			[^ #()]
		ifFalse:
			[process := self tryEvaluatingAsSmalltalk: aCommandString.
			(process exitStatus == #success)
				ifTrue: "aCommandString was successfully evaluated as a Smalltalk expression"
					[self outputStream nextPutAll: process upToEnd.
					self errorStream nextPutAll: process errorUpToEnd.
					self saveInHistoryList: aCommandString.
					^ #()]
				ifFalse: "Treat aCommandString as a command pipeline"
					[^ self pipelineCollectionFrom: aCommandString]]
! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 12/11/2007 18:39'!
evaluatePipeline: aPipeline
	"Evaluate aPipeline with event handlers to process output and error data."

	"CommandShell new evaluatePipeline: 
			(CommandShell new
				pipelineCollectionFrom: 'ls NOSUCHFILE * | cat | dd | cat | wc; ps -aef | cat& pwd | cat') first"

	| index |
	self registerOutputEventsFor: aPipeline.
	self registerErrorEventsFor: aPipeline.
	aPipeline background
		ifTrue:
			[index := self registerBackgroundPipeline: aPipeline.
			self registerBackgroundCompletionEventsFor: aPipeline
				commandIndex: index.
			self backgroundProxies addAll: aPipeline.
			aPipeline value.
			"Print the index of the background process."
			self errorStream nextPutAll: '[', index asString, '] '.
			aPipeline last isExternalProcess ifTrue:
				[aPipeline last processProxy pid printOn: self errorStream].
			self errorStream cr; nextPut: Character lf]
		ifFalse:
			[self keyboardConnector: aPipeline pipeToInput.
			self registerForegroundCompletionEventsFor: aPipeline.
			self foregroundProxies addAll: aPipeline.
			aPipeline value].
	^ aPipeline
! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 9/3/2010 17:04'!
evaluatePipelines: aPipelineCollection
	"Evaluate the process proxies in aPipelineCollection, and process their output and error
	streams. Answer aPipelineCollection."

	| lastForegroundPipeline completionSemaphore |
	lastForegroundPipeline := nil.
	aPipelineCollection isEmpty ifFalse:
		[lastForegroundPipeline := aPipelineCollection reversed
			detect: [:p | p background not]
			ifNone: [].
		lastForegroundPipeline ifNotNil: [self registerPromptEventFor: lastForegroundPipeline].
		aPipelineCollection do: [:pipeline |
			completionSemaphore := Semaphore new.
			pipeline background
				ifTrue: [completionSemaphore signal "no wait"]
				ifFalse: [pipeline when: #complete send: #signal to: completionSemaphore].
			self evaluatePipeline: pipeline.
			completionSemaphore wait]].
	lastForegroundPipeline ifNil: [self notifyPrompt].
	self resetSafeToExitFlag.
	^ aPipelineCollection
! !

!CommandShell methodsFor: 'process proxy creation' stamp: 'dtl 11/21/2006 09:21'!
evaluationProxyFor: aCommandString input: in output: out error: err
	"Answer a PipeableEvaluator on a doIt command, or nil."

	"CommandShell new evaluationProxyFor: 'Smalltalk' input: nil output: nil error: nil"
	"CommandShell new evaluationProxyFor: 'BOGUS' input: nil output: nil error: nil"

	| proxy |
	proxy := ExpressionEvaluator
				command: aCommandString
				pipeFrom: in
				pipeTo: out
				errorPipelineStream: err.
	^ proxy
! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 6/9/2001 16:13'!
filterAndForward: aCharacter
	"Filter aCharacter, taking special action if needed. If a child process is active,
	forward aCharacter to the child and answer nil. Otherwise answer aCharacter."

	^ (self checkSttyFor: aCharacter asciiValue) ifNotNil:
		[keyboardConnector isNil
			ifTrue:
				[aCharacter]
			ifFalse:
				[keyboardConnector closed
					ifTrue:
						[aCharacter]
					ifFalse:
						[keyboardConnector nextPut: aCharacter; flush.
						nil]]]! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 11/7/2007 07:01'!
filterAndForwardEvent: keyboardEvent 
	"Filter keyboardEvent, taking special action if needed. If a child process
	is active, forward keyboardEvent to the child and answer nil. Otherwise
	answer keyboardEvent."

	(keyboardConnector isNil or: [keyboardConnector closed])
		ifTrue: [^ keyboardEvent]
		ifFalse: [keyboardConnector
					nextPut: keyboardEvent keyValue asCharacter;
					flush.
			^ nil]! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 7/7/2006 09:19'!
foregroundPipelineComplete: aPipeline
	"This message is sent when aPipeline has completed evaluation.
	Ensure that the output and error event handlers have had a chance
	to completely empty the pipes prior to cleaning things up."

	self waitUntilPipesAreEmpty: aPipeline.
	aPipeline closePipes; release.
	self keyboardConnector: nil.
	[self foregroundProxies removeAll: aPipeline]
		on: Error
		do: ["exit command may cause the collection to be eliminated, hence an error"]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/9/2006 10:23'!
foregroundProxies
	"All of my process proxies that are being evaluated in the foreground."

	^ foregroundProxies ifNil: [foregroundProxies := OrderedCollection new]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/9/2006 10:23'!
foregroundProxies: aCollection
	"All of my process proxies that are being evaluated in the foreground."

	foregroundProxies := aCollection
! !

!CommandShell methodsFor: 'command history' stamp: 'dtl 4/22/2001 10:48'!
historyAt: anInteger

	| bufferSize index |
	bufferSize := commandHistory size / 2.
	(anInteger > 0)
		ifTrue:
			[index := bufferSize + 1 - historyIndex + anInteger * 2]
		ifFalse:
			[index := bufferSize + anInteger * 2].
	((index > commandHistory size) | (index < 1))
		ifTrue: [^ '']
		ifFalse: [^ commandHistory at: index]
! !

!CommandShell methodsFor: 'defaults' stamp: 'dtl 4/8/2001 15:26'!
historyDepth
	"Number of commands to retain in the command history list"

	^ 100! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 4/3/2001 22:48'!
historyIndex

	historyIndex ifNil: [historyIndex := 1].
	^ historyIndex
! !

!CommandShell methodsFor: 'working directory' stamp: 'dtl 12/29/2002 20:14'!
home
	"Home directory. Use the environment variable, and set it in #initialize."

	^ self environment at: #HOME
! !

!CommandShell methodsFor: 'command scripting' stamp: 'dtl 3/18/2003 23:12'!
if: aCommandStringOrScript then: successBlock else: failureBlock
	"Evaluate aCommandStringOrScript, then evaluate successBlock or failureBlock
	depending on the exit status of the command or script. The blocks may use
	either zero or one argument. If a single argument is used, the collection of
	evaluated pipelines is passed to the block as an argument."

	"CommandShell new open;
		if: 'who'
		then: ['the command succeeded']
		else: ['the command failed']"
	"CommandShell new
		if: 'who
			ls NOSUCHFILE
			pwd'
		then: ['the script succeeded']
		else: ['the script failed']"
	"CommandShell new
		if: 'who
			ls NOSUCHFILE
			pwd'
		then: [:p | 'the script succeeded ', p printString]
		else: [:p | 'the script failed ', p printString]"

	| pipelines |
	pipelines := self script: aCommandStringOrScript
		onFailureDo: [:p | (failureBlock numArgs == 0)
						ifTrue: [^ failureBlock value]
						ifFalse: [^ failureBlock value: p]].
	(successBlock numArgs == 0)
		ifTrue: [^ successBlock value]
		ifFalse: [^ successBlock value: pipelines]
! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/29/2002 20:00'!
initialize

	self installCommands.
	self setHomeDirectory
! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/27/2001 12:24'!
installCommand: aShellBuiltin

	self builtinCommands at: aShellBuiltin name put: aShellBuiltin! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 7/14/2002 12:05'!
installCommands

	self installCommands: self defaultBuiltinCommands
! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 7/14/2002 12:05'!
installCommands: aCommandArray

	self builtinCommands: (ShellBuiltin commandDictionaryFor: self commands: aCommandArray)
! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 6/7/2001 23:17'!
interruptCharacterValue
	"<ctl>-C"

	^ 3! !

!CommandShell methodsFor: 'defaults' stamp: 'dtl 4/8/2001 15:11'!
invalidCommandMessage

	^ 'invalid command', Character cr asString! !

!CommandShell methodsFor: 'testing' stamp: 'dtl 9/26/2009 13:15'!
isBuiltInCommand: aCommandString
	"Answer true if aCommandString can be evaluated as a built in command."

	"CommandShell new initialize isBuiltInCommand: 'help this should answer true'"
	"CommandShell new initialize isBuiltInCommand: 'this should answer false'"

	^ self builtinCommands includesKey: (aCommandString copyUpTo: Character space)! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 6/7/2001 23:20'!
isEndOfFile: anInteger

	^ anInteger == self endOfFileCharacterValue
! !

!CommandShell methodsFor: 'testing' stamp: 'dtl 4/7/2001 13:04'!
isExcludedFromHistory: commandString
	"Answer true if commandString should not be remembered in the history list. The 'fc'
	and 'history' commands should be excluded."

	^ ('fc*' match: commandString)
		or: ['history*' match: commandString]! !

!CommandShell methodsFor: 'testing' stamp: 'dtl 4/19/2003 09:13'!
isExternalProcess
	"Answer true if the process which I represent is an external OSProcess. For protocol
	compatibility with PipeJunction."

	^ false! !

!CommandShell methodsFor: 'input character filtering' stamp: 'dtl 6/7/2001 23:21'!
isInterrupt: anInteger

	^ anInteger == self interruptCharacterValue
! !

!CommandShell methodsFor: 'testing' stamp: 'dtl 1/20/2007 14:15'!
isOneWordCommentString: aCommandString
	"Answer true if aCommandString is one word beginning with $#, or if it is an empty string."

	"CommandShell new isOneWordCommentString: '#oneWordComment'"
	"CommandShell new isOneWordCommentString: ' #oneWordComment'"
	"CommandShell new isOneWordCommentString: ' #oneWordComment '"
	"CommandShell new isOneWordCommentString: '#three word comment'"
	"CommandShell new isOneWordCommentString: 'not a comment'"
	"CommandShell new isOneWordCommentString: 'notAComment'"
	"CommandShell new isOneWordCommentString: ''"

	| s |
	s := aCommandString withBlanksTrimmed.
	^ s isEmpty
		or: [(s first == self shellSyntax commentChar)
			and: [(Scanner new scanTokens: s) size == 1]]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 4/22/2001 18:49'!
keyboardConnector
	"If nil, then my views should behave normally. If a WriteStream, then my
	views should attempt to forward all keyboard inputs to the WriteStream.
	This is to permit the keyboard to be connected to an externally executing
	process, then reconnected to its normal functions when the external process
	is complete."


	^ keyboardConnector! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 4/22/2001 18:56'!
keyboardConnector: aWriteStreamOrNil
	"If nil, then my views should behave normally. If a WriteStream, then my
	views should attempt to forward all keyboard inputs to the WriteStream.
	This is to permit the keyboard to be connected to an externally executing
	process, then reconnected to its normal functions when the external process
	is complete."

	keyboardConnector := aWriteStreamOrNil
! !

!CommandShell methodsFor: 'VM building' stamp: 'dtl 8/7/2002 11:50'!
makeVmIn: buildDirectoryPathName
	"Rebuild the virtual machine and plugins in the buildDirectoryPathName
	directory. If the build is successful, save the image and restart using the
	new VM. This assumes that the currently executing VM is either located in,
	or linked to, the buildDirectoryPathName directory."

	"CommandShell makeVmIn: FileDirectory default pathName, FileDirectory slash, 'build'"

	self canAccessOSProcess
		ifFalse:
			[^ self notify: 'OSProcess not available or OSProcess plugin not functioning'].
	self if: 'cd ', buildDirectoryPathName
		then:
			[self if: 'make'
				then:
					[self script: 'echo make completed successfully, restarting VM; OSProcess quitAndRestart !!']
				else:
					[self script: 'echo make failed']]
		else:
			[self script: 'echo cannot change directory to ', buildDirectoryPathName]
! !

!CommandShell methodsFor: 'command prompt' stamp: 'dtl 1/20/2007 10:49'!
notifyAlternatePrompt
	"Display $PS2"

	self changed: #alternatePrompt! !

!CommandShell methodsFor: 'command prompt' stamp: 'dtl 1/20/2007 10:48'!
notifyPrompt
	"Display $PS1"

	self changed: #prompt! !

!CommandShell methodsFor: 'command prompt' stamp: 'dtl 1/20/2007 10:49'!
notifyRestorePrompt
	"Restore last prompt display"

	self changed: #restorePrompt! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 1/20/2007 09:26'!
open

	| tty |
	tty := CommandShellTranscript commandShell: self.
	tty open.
	^ tty

! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/11/2007 18:46'!
openLabel: aString 

	| tty |
	tty := CommandShellTranscript commandShell: self.
	tty openLabel: aString.
	^ tty

! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 1/16/2007 07:00'!
outputStream
	"Answer the value of outputStream"

	^ outputStream ifNil: [outputStream := WriteStream on: '']! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 8/19/2006 13:14'!
outputStream: anObject
	"Set the value of outputStream"

	outputStream := anObject! !

!CommandShell methodsFor: 'process proxy creation' stamp: 'dtl 10/13/2008 21:22'!
pipeableProxyFor: aCommandString input: in output: out error: err predecessorProxy: lastProxy
	"Answer a proxy for an external OS process, or for an internal builtin command,
	or for an evaluation string. The proxy is suitable for use in a command pipeline,
	possibly mixing internal and external commands."

	"(CommandShell new pipeableProxyFor: '#''this is a doIt'' !! ' input: nil output: nil error: nil predecessorProxy: nil) value upToEnd"
	"(CommandShell new pipeableProxyFor: 'help' input: nil output: nil error: nil predecessorProxy: nil) value upToEnd"
	"(CommandShell new pipeableProxyFor: 'cat /etc/hosts' input: nil output: nil error: nil predecessorProxy: nil) value upToEnd"

	| errorPipelineStream cmd |
	errorPipelineStream := lastProxy ifNotNil: [lastProxy errorPipeForNextProxy].
	^ (((self doItProxyFor: aCommandString						"try entire line as a Smalltalk expression"
			input: in
			output: out
			error: (err ifNil: [errorPipelineStream]))
		ifNil: [self builtinCommandProxyFor: aCommandString	"is it a shell builtin?"
				input: in
				output: out
				error: (err ifNil: [errorPipelineStream])])
			ifNil: [self processProxyFor: aCommandString			"then it must be an external command"
					input: in
					output: out
					error: err
					predecessorProxy: lastProxy])
				ifNil: [cmd := (aCommandString findTokens: ' ') first.		"otherwise create evaluator to print error message"
					self evaluationProxyFor: 'stderr nextPutAll: ''sqsh: ', cmd, ': command not found''; cr. self fail'
						input: in
						output: out
						error: err]
! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 11/20/2006 08:10'!
pipeline: aCommand
	"Find the individual commands in a command pipeline string, and execute them
	as a command pipeline. Answer a ProxyPipeline collection of processes in the pipeline."

	"(CommandShell new pipeline: 'ps | grep ps | wc') last output"
	"CommandShell new pipeline: 'ls NOSUCHFILE * | dd | cat'"

	| sema pipeline |
	sema := Semaphore new.
	pipeline := ProxyPipeline fromString: aCommand shell: self.
	pipeline when: #complete send: #signal to: sema.
	pipeline value.
	sema wait.
	^ pipeline
! !

!CommandShell methodsFor: 'command line parsing' stamp: 'dtl 1/20/2007 14:12'!
pipelineCollectionFrom: aCommandString
	"Break aCommandString into individual pipeline command strings, and answer a
	collection of process proxy pipelines."

	"CommandShell new pipelineCollectionFrom: 'who | cat | wc; ps -aef | cat& pwd | cat'"

	| str pipelineCollection tokens strm cmd cmdWithTerm pipeline lastProxy |
	str := aCommandString withBlanksTrimmed.
	pipelineCollection := OrderedCollection new.
	(self shellSyntax isComment: str)
		ifFalse:
			[tokens := str findTokens: ';&' keep: ';&'.
			strm := ReadStream on: tokens.
			[strm atEnd]
				whileFalse:
					[cmd := strm next withBlanksTrimmed.
					(self shellSyntax isComment: cmd)
						ifFalse:
							[strm atEnd
								ifTrue:
									[((cmd size == 1) and: [((cmd at: 1) == $;) | ((cmd at: 1) == $&)])
										ifTrue:
											["Special case, user has entered ';<cr>' or '&<cr>'"
											cmd := cmdWithTerm := '']
										ifFalse:
											[cmdWithTerm := cmd]]
								ifFalse:
									[cmdWithTerm := cmd, strm peek].
							self saveInHistoryList: cmdWithTerm.
							(cmdWithTerm size > 0)
								ifTrue:
									[pipeline := ProxyPipeline fromString: cmd shell: self.
									pipeline background:
										(strm next = self shellSyntax noWaitToken).
									lastProxy := pipeline last.
									lastProxy ifNil:
										[self inform: 'cannot access system'.
										^ nil].
									lastProxy outputPipeForNextProxy
										ifNotNil:
											[lastProxy setNonBlockingOutput].
									pipelineCollection add: pipeline]]]].
	^ pipelineCollection
! !

!CommandShell methodsFor: 'defaults' stamp: 'dtl 6/4/2006 18:01'!
pollingDelayTime
	"For various polling loops, number of milliseconds to delay"

	^ 50
! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 1/13/2007 17:55'!
processCommand: aCommandString
	"Evaluate aCommandString in a separate Smalltalk process. This permits the
	user interface to remain responsive."

	self commandLineProcesses add:
		([self command: aCommandString.
		self commandLineProcesses
				remove: Processor activeProcess
				ifAbsent: []] forkAt: Processor userBackgroundPriority)
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 11/23/2002 13:31'!
processProxies
	"All of my process proxies."

	^ self foregroundProxies, self backgroundProxies
! !

!CommandShell methodsFor: 'process proxy creation' stamp: 'dtl 10/14/2008 20:11'!
processProxyFor: aCommandString input: in output: out error: err predecessorProxy: lastProxy
	"Answer a proxy for an external OS command process. Note that the sender is
	responsible for closing the input pipe. Under certain conditions, reading up to
	end on a proxy can deadlock the VM if the input pipe remains open."

	"CommandShell new processProxyFor: '/bin/sh' input: nil output: nil error: nil predecessorProxy: nil"

	| input newInputPipe errorPipelineStream p proxy |
	self canAccessOSProcess
		ifFalse:
			["Answer a doIt proxy with an error message. Call it this way to make sure that
			the error streams get wired up correctly"
			p := self pipeableProxyFor: 'String new: 0 !! '
					input: in
					output: out
					error: err
					predecessorProxy: lastProxy.
			p errorPipelineStream nextPutAll:
				'cannot access system to run ''', aCommandString, '''', Character cr asString.
			^ p].
	((in isNil or: [lastProxy isNil]) or: [in isPipe not])
		ifTrue:
			[input := in]
		ifFalse:
			["Input from a command pipeline"
			(in isKindOf: InternalPipe)
				ifTrue:
					["The input stream is not an external OS pipe. Need to create one,
					and move the contents of the in stream into the new OSPipe prior
					to starting the external OS process."
					newInputPipe := PipeJunction externalPipeClass blockingPipe.
					lastProxy replaceOutputStreamWith: newInputPipe writer.
					input := newInputPipe reader]
				ifFalse:
					[input := in reader]].
	errorPipelineStream := lastProxy ifNotNil: [lastProxy errorPipeForNextProxy].
	proxy := PipeableOSProcess
		commandNoEvaluate: aCommandString
		environment: self environment
		workingDir: self workingDirectory
		input: input
		output: out
		error: err
		errorPipelineStream: errorPipelineStream
		shellSyntax: self shellSyntax.
	(proxy isNil and: [newInputPipe notNil])
		ifTrue: [newInputPipe closeReader]. "replacement pipe will not be read, so close its output"
	^ proxy
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/17/2002 23:45'!
promptString
	"The primary command prompt string, equivalent to $PS1 in a Unix shell"

	^ promptString ifNil: [promptString := self defaultPromptString]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/17/2002 23:47'!
promptString: aString
	"The primary command prompt string, equivalent to $PS1 in a Unix shell"

	promptString := aString
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/17/2002 23:46'!
promptStringTwo
	"The secondary command prompt string, equivalent to $PS2 in a Unix shell"

	^ promptStringTwo ifNil: [promptStringTwo := self defaultPromptStringTwo]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/17/2002 23:48'!
promptStringTwo: aString
	"The secondary command prompt string, equivalent to $PS2 in a Unix shell"

	promptStringTwo := aString
! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 1/20/2007 13:41'!
promptWhenComplete: aPipeline

	self waitUntilPipesAreEmpty: aPipeline.
	self notifyPrompt! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/25/2001 23:07'!
readSyncSemaphore
	"Used in MVC to pause the screen output when presenting a dialogue to the user"

	readSyncSemaphore ifNil: [readSyncSemaphore := Semaphore forMutualExclusion].
	^ readSyncSemaphore! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 6/25/2001 20:34'!
readSyncSemaphore: aSemaphore
	"Used in MVC to pause the screen output when presenting a dialogue to the user"

	readSyncSemaphore := aSemaphore! !

!CommandShell methodsFor: 'process proxy creation' stamp: 'dtl 4/8/2006 18:18'!
redirectedPipeableProxyFor: aCommandString predecessorProxy: lastProxy
	"Answer a proxy for an external OS process, or for an internal builtin command,
	or for an evaluation string. The proxy is suitable for use in a command pipeline,
	possibly mixing internal and external commands."

	"(CommandShell new redirectedPipeableProxyFor: '#''this is a doIt'' !! '
							predecessorProxy: nil) value upToEnd"
	"(CommandShell new redirectedPipeableProxyFor: 'help'
							predecessorProxy: nil) value upToEnd"
	"(CommandShell new redirectedPipeableProxyFor: 'cat /etc/hosts'
							predecessorProxy: nil) value upToEnd"

	| redirectedCommand nextProxy input |
	redirectedCommand := self shellSyntax
								redirectIOFor: aCommandString
								inDirectoryPath: self workingDirectory.
	lastProxy isNil
		ifTrue:
			[input := redirectedCommand at: 2]
		ifFalse:
			[(redirectedCommand at: 2) isNil
				ifTrue:
					[input := lastProxy outputPipeForNextProxy]
				ifFalse:
					[lastProxy replaceOutputStreamWith: nil.
					input := redirectedCommand at: 2]].
	nextProxy := self pipeableProxyFor: (redirectedCommand at: 1)
		input: input
		output: (redirectedCommand at: 3)
		error: (redirectedCommand at: 4)
		predecessorProxy: lastProxy.
	^ nextProxy handleRedirectedOutputOf: lastProxy
! !

!CommandShell methodsFor: 'event registration' stamp: 'dtl 1/20/2007 13:46'!
registerBackgroundCompletionEventsFor: aPipeline commandIndex: index
	"When an event driven pipeline is complete, it will trigger #complete."

	aPipeline when: #complete
		send: #backgroundPipelineComplete:commandIndex:
		to: self
		withArguments: {aPipeline . index}

! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 12/15/2002 16:48'!
registerBackgroundPipeline: aPipeline
	"Register aPipeline and answer an index number identifier."

	| index |
	aPipeline background
		ifTrue:
			["Find available slot in registry, expanding if needed"
			index := self backgroundPipelines identityIndexOf: nil.
			(index > 0)
				ifTrue:
					[self backgroundPipelines at: index put: aPipeline.
					^ index]
				ifFalse:
					[self backgroundPipelines addLast: aPipeline.
					^ backgroundPipelines size]]
		ifFalse:
			[self error: 'expected a background pipeline']
! !

!CommandShell methodsFor: 'event registration' stamp: 'dtl 1/13/2007 17:22'!
registerErrorEventsFor: aPipeline
	"When an event driven PipeJunction has error data available, it will
	trigger #errorDataReady."

	aPipeline when: #errorDataReady
		send: #copyAllErrorFrom:
		to: self
		with: aPipeline
! !

!CommandShell methodsFor: 'event registration' stamp: 'dtl 12/15/2002 13:29'!
registerForegroundCompletionEventsFor: aPipeline
	"When an event driven pipeline is complete, it will trigger #complete."

	aPipeline when: #complete
		send: #foregroundPipelineComplete:
		to: self
		with: aPipeline
! !

!CommandShell methodsFor: 'event registration' stamp: 'dtl 1/13/2007 17:22'!
registerOutputEventsFor: aPipeline
	"When an event driven PipeJunction has new data available, it will
	trigger #outputDataReady."

	aPipeline when: #outputDataReady
		send: #copyAllOutputFrom:
		to: self
		with: aPipeline
! !

!CommandShell methodsFor: 'event registration' stamp: 'dtl 1/20/2007 13:45'!
registerPromptEventFor: aPipeline

	aPipeline when: #complete
		send: #promptWhenComplete:
		to: self
		with: aPipeline
! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 6/24/2002 07:02'!
release

	self startUpInNewSession.
	^ super release! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 6/4/2006 17:44'!
resetSafeToExitFlag
	"When all foreground and background proxies are complete, the safeToExit
	flag may be set to true. The flag is set false when a snaphot command is
	evaluated, and is checked by any command which would cause Squeak to
	exit. This prevents Squeak from automatically exiting when restarting the
	image."

	[[self foregroundProxies isEmpty not or: [self backgroundProxies isEmpty not]]
		whileTrue: [(Delay forMilliseconds: self pollingDelayTime * 2) wait].
	self safeToExit: true] forkAt: Processor userBackgroundPriority
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 7/14/2002 15:56'!
safeToExit
	"This is a guard variable to prevent the user from running a command line containing both
	a snapshot command and an exit command. This prevents the user from entering a command
	such as 'snapshot; exit' which would cause the image to immediately exit when next restarted."

	^ safeToExit ifNil: [safeToExit := true]
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 7/14/2002 15:56'!
safeToExit: trueOrFalse
	"This is a guard variable to prevent the user from running a command line containing both
	a snapshot command and an exit command. This prevents the user from entering a command
	such as 'snapshot; exit' which would cause the image to immediately exit when next restarted."

	safeToExit := trueOrFalse
! !

!CommandShell methodsFor: 'command history' stamp: 'dtl 4/7/2001 13:05'!
saveInHistoryList: aCommandString

	((aCommandString size == 0) or: [self isExcludedFromHistory: aCommandString])
		ifFalse:
			[self commandHistory add: self historyIndex printString; add: aCommandString.
			historyIndex := historyIndex + 1]
! !

!CommandShell methodsFor: 'command support' stamp: 'dtl 6/29/2010 21:22'!
scheduleToEvaluate: aBlock
	"Evaluate aBlock, typically to create a new scheduled window. Make it work in
	both Morphic and MVC. In Morphic, just evaluate aBlock, but in MVC, put it in
	a queue for evaluation within a control loop. This method may be sent from
	a process running independent of MVC controller scheduling.
	
	Newer Squeak images implement #addDeferredUIMessage in the current
	project, eliminating the need for an #isMorphic test. This mechanism is not
	available for older images or for Pharo."

	CommandShell isMorphic
		ifTrue: [WorldState addDeferredUIMessage: aBlock]
		ifFalse: [ScheduledControllers activeController addDeferredUIMessage: aBlock]! !

!CommandShell methodsFor: 'command scripting' stamp: 'dtl 12/15/2002 18:50'!
script: aScriptString
	"A script is a String containing a list of commands separated by Character cr.
	Evaluate each command string in the array until a failure is encountered.
	Answer the last collection of pipelines to have been evaluated."

	"CommandShell new open; 
		script: 'who
				help
				ls -l NOSUCHFILE
				help'"

	^ self script: aScriptString onFailureDo: [:p | p]

! !

!CommandShell methodsFor: 'command scripting' stamp: 'dtl 1/21/2007 20:19'!
script: aScriptString onFailureDo: aBlockWithZeroOrOneParameter
	"A script is a String containing a list of commands separated by Character cr.
	Evaluate each command in aScriptString until a failure is encountered.
	On failure, answer the result of evaluating aBlockWithOneParameter with
	the collection of evaluated pipelines as its parameter. Otherwise, answer the
	last collection of pipelines to have been evaluated."

	"CommandShell new open; 
		script: 'who
				help
				ls -l NOSUCHFILE
				help'
		onFailureDo: [:p | 'the pipeline collection that failed was ', p printString]"
	"CommandShell new open; 
		script: 'who
				help
				ls
				help'
		onFailureDo: [:p | 'the pipeline collection that failed was ', p printString]"
	"CommandShell new open; 
		script: 'who
				help
				ls -l NOSUCHFILE
				help'
		onFailureDo: ['the pipeline collection failed']"

	| commandStream pipelines |
	commandStream := ReadStream on:
						(aScriptString
							copyReplaceAll: Character lf asString
							with: Character cr asString).
	pipelines := nil.
	[commandStream atEnd]
		whileFalse:
			[pipelines := self
				command: commandStream nextLine
				onFailureDo:
					[:p | (aBlockWithZeroOrOneParameter numArgs == 0)
						ifTrue: [^ aBlockWithZeroOrOneParameter value]
						ifFalse: [^ aBlockWithZeroOrOneParameter value: p]]].
	^ pipelines

! !

!CommandShell methodsFor: 'command scripting' stamp: 'dtl 12/15/2002 18:47'!
scriptDoAll: aScriptString
	"A script is a String containing a list of commands separated by Character cr.
	Evaluate each command in aScriptString. Ignore failures and force evaluation of
	all commands in the script. Answer the last pipeline collection to have been
	evaluated."

	"CommandShell new open; 
		scriptDoAll: 'who
				help
				ls -l NOSUCHFILE
				help'"

	| commandStream lastPipelineCollection |
	commandStream := ReadStream on:
						(aScriptString
							copyReplaceAll: Character lf asString
							with: Character cr asString).
	[commandStream atEnd]
		whileFalse:
			[lastPipelineCollection := self command: commandStream nextLine].
	^ lastPipelineCollection

! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/29/2002 20:08'!
setHomeDirectory

	self environment.
	^ environment at: #HOME put: FileDirectory default pathName
! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 12/17/2001 11:34'!
shellSyntax
	"A pluggable instance of a command processing syntax"

	^ shellSyntax ifNil: [shellSyntax := ShellSyntax new]! !

!CommandShell methodsFor: 'accessing' stamp: 'dtl 9/8/2001 11:10'!
shellSyntax: aSyntax
	"A pluggable instance of a command processing syntax"

	shellSyntax := aSyntax! !

!CommandShell methodsFor: 'command support' stamp: 'dtl 1/20/2007 14:09'!
splitPipelineCommands: aCommand
	"Answer a list of the individual commands in a command pipeline string,
	delimited by the pipe character."

	^ self shellSyntax splitPipelineCommands: aCommand
! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/31/2002 11:53'!
startUpInNewSession
	"Do this if the image has been restarted, in which case all external process
	references are invalid."

	self clearSemaphore.
	self terminateCommandLineProcesses.
	foregroundProxies := backgroundProxies := backgroundPipelines := nil.
	self keyboardConnector: nil
! !

!CommandShell methodsFor: 'initialize-release' stamp: 'dtl 12/15/2002 17:33'!
terminateCommandLineProcesses

	(self commandLineProcesses reject: [:e | e == Processor activeProcess])
		do: [:p | p terminate].
	commandLineProcesses := nil
! !

!CommandShell methodsFor: 'command line parsing' stamp: 'dtl 4/20/2003 15:35'!
tryEvaluatingAsSmalltalk: aCommandString
	"Attempt to treat aCommandString as a Smalltalk expression running in an
	evaluation process. Evaluate the proxy and set exit status to reflect failure
	on either the expression compilation or the runtime evaluation. Answer the
	process proxy."

	"CommandShell new tryEvaluatingAsSmalltalk: 'this expression should fail in the compiler'"
	"CommandShell new tryEvaluatingAsSmalltalk: '#this expressionShouldFailInRuntimeEvaluation'"
	"CommandShell new tryEvaluatingAsSmalltalk: '#thisIsAValidSymbolExpression'"
	"CommandShell new tryEvaluatingAsSmalltalk: 'Smalltalk'"
	"CommandShell new tryEvaluatingAsSmalltalk: '#oneWordCommentShouldFail'"

	| process |
	process := self evaluationProxyFor: aCommandString input: nil output: nil error: nil.
	process value.
	process closeErrorPipeline. "Last process in pipeline, so it can be closed."
	^ process! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 12/15/2002 16:34'!
unRegisterBackgroundPipeline: aPipeline at: index

	(self backgroundPipelines at: index) == aPipeline
		ifTrue: [self backgroundPipelines at: index put: nil]
		ifFalse: [self error: 'inconsistent registry']
! !

!CommandShell methodsFor: 'evaluation' stamp: 'dtl 1/13/2007 18:09'!
waitForCommand: aCommandString
	"Evaluate aCommandString in a separate Process, and signal completionSemaphore
	when complete. Answer a collection of evaluated pipelines, or nil if nothing was done."

	| sema pipelines lastPipeline |
	self show: aCommandString; cr.
	pipelines := self evaluateOrMakePipelinesFrom: aCommandString.
	pipelines isEmpty ifFalse:
		[lastPipeline := pipelines last.
		lastPipeline isComplete ifFalse:
			[sema := Semaphore new.
			lastPipeline when: #complete
				send: #signal
				to: sema.
			[self evaluatePipelines: pipelines] forkAt: Processor userBackgroundPriority.
			lastPipeline isComplete ifFalse: [sema wait].
			self waitUntilPipesAreEmpty: lastPipeline]].
	^ pipelines
! !

!CommandShell methodsFor: 'event handling' stamp: 'dtl 10/3/2006 07:21'!
waitUntilPipesAreEmpty: aPipeline
	"Some other Smalltalk process may be updating the pipes, or we may be
	waiting for data in an external OS pipe to be read into the image. Wait
	for these to reach completion."

	[aPipeline pipesAreEmpty] whileFalse: [(Delay forMilliseconds: 100) wait]
! !

!CommandShell methodsFor: 'working directory' stamp: 'dtl 12/16/2001 20:27'!
workingDirectory

	^ self shellSyntax workingDirectory
		ifNil:
			[shellSyntax workingDirectory: FileDirectory default pathName; workingDirectory]! !

!CommandShell methodsFor: 'working directory' stamp: 'dtl 12/16/2001 20:15'!
workingDirectory: aPathString

	self shellSyntax workingDirectory: aPathString
! !

Object subclass: #ShellSyntax
	instanceVariableNames: 'cwdDictionary cwdIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Base'!

!ShellSyntax commentStamp: 'dtl 9/7/2009 12:36' prior: 0!
My instances implement parsing of strings in a manner similar to a simple Unix command shell. I provide path name expansion in the context of an external file system, and support the syntax required for IO redirection. All file name globbing and PATH searching are implemented in Smalltalk rather than in C library functions or an external command shell.

Most of my syntax is applicable for any operating system. Where needed, platform specific methods are in my "platform dependent" category. Currently, Unix and Windows are supported, and other platforms have not been tested. The primary difference between Unix and Windows support is that device names (such as 'C:') are used in Windows path strings. Separate current working directory strings are maintained for all Windows device names. For Unix, a single current working directory path is used. On Windows, this permits the CommandShell 'cd' command to support changing directories to another device without losing track of the current working directory for the previous device.

Command pipeline syntax is not supported here. See CommandShell for the implementation of command pipelines.
!

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 5/2/2002 00:12'!
appendPath: aPathString toPath: startingPathString 
	"ShellSyntax new appendPath: 'hosts' toPath: '/etc'"
	"ShellSyntax new appendPath: 'etc' toPath: '/'"
	"ShellSyntax new appendPath: 'hosts' toPath: '/etc'"
	"ShellSyntax new appendPath: 'WINDOWS' toPath: 'C:\'"
	"ShellSyntax new appendPath: 'someFileName' toPath: 'C:\WINDOWS'"

	| newPath |
	^ (aPathString findTokens: FileDirectory slash)
		inject: startingPathString
		into: [:dirPath :nextPart | 
			nextPart = '..'
				ifTrue: [newPath := self upOneDirectoryLevel: dirPath]
				ifFalse: [nextPart = '.'
						ifTrue: [newPath := dirPath]
						ifFalse: [newPath := (self isAtFileSystemRoot: dirPath)
										ifTrue: [dirPath , nextPart]
										ifFalse: [dirPath , FileDirectory slash , nextPart]]].
			newPath]! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 5/11/2002 20:15'!
argsFromCommand: aString
	"Answer a collection of strings from a whitespace delimited string. Single quote characters
	protect substrings from expansion to tokens. Single quote characters may be escaped by
	the backslash character to prevent this interpretation."

	^ (self stringTokensFrom: aString) first
! !

!ShellSyntax methodsFor: 'private' stamp: 'dtl 8/7/2002 13:09'!
canAccessOSProcess
	"True if OSProcess is in the image and if the OSProcess plugin is functioning properly"

	"ShellSyntax new canAccessOSProcess"

	| osp |
	^ (osp := Smalltalk at: #OSProcess ifAbsent: []) notNil and: [osp accessor canAccessSystem]
! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 3/10/2001 18:07'!
character: aCharacter matches: aToken
	"Test if aCharacter is included in the characters of aToken, or if aToken is a #matchAny"

	(aToken == #matchAny) ifTrue: [^ true].
	^ aToken includes: aCharacter! !

!ShellSyntax methodsFor: 'defaults' stamp: 'dtl 9/8/2001 10:15'!
commentChar

	^ $#! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 9/12/2002 13:19'!
containsExpansionTokens: aString
	"Answer true if aString contains wildcard expansion characters, or if it contains
	anything that might be construed as a regular expression for globbing, or if it
	contains path separator characters. If false, the sender may safely skip normal
	glob expansion on aString."

	^ aString includesAnyOf: {$? . $* . $[ . FileDirectory pathNameDelimiter}
! !

!ShellSyntax methodsFor: 'accessing' stamp: 'dtl 12/16/2001 20:04'!
cwdDictionary
	"Current working directory paths indexed by device or file system name"

	^ cwdDictionary ifNil: [cwdDictionary := Dictionary new]
! !

!ShellSyntax methodsFor: 'accessing' stamp: 'dtl 12/17/2001 10:49'!
cwdIndex
	"Key to current cwd path name in cwdDictionary"

	^ cwdIndex ifNil: [cwdIndex := self cwdIndexFor: FileDirectory default pathName]
! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 12/16/2001 20:06'!
cwdIndex: aSymbol
	"Key to current cwd path name in cwdDictionary"

	cwdIndex := aSymbol
! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 12/17/2001 09:56'!
cwdIndexFor: aPathStringWithPrefix
	"Answer key to use for aPathStringWithPrefix"

	| prefix |
	prefix := (self splitLocationPrefixFromPath: aPathStringWithPrefix) first.
	^ prefix isEmpty
		ifTrue: [#default]
		ifFalse: [prefix]! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 1/3/2002 19:15'!
defaultPathFor: aPathStringWithPrefix
	"Answer a reasonable path string to use as a default"

	^ self pathRoot: aPathStringWithPrefix
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 1/21/2007 14:01'!
doItStringFrom: aCommandString
	"Answer a string which can be evaluated as Smalltalk doIt, or nil."

	| s |
	s := aCommandString withoutTrailingBlanks.
	s last == $!!
		ifTrue: [^ aCommandString copyUpToLast: $!!]
		ifFalse: [^ nil]! !

!ShellSyntax methodsFor: 'defaults' stamp: 'dtl 4/28/2002 16:09'!
escapeChar
	"The character to use to escape the normal interpretation of the following character
	in a string."

	^ $\! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 12/28/2002 11:36'!
expandArgument: aString inDirectoryPath: aDefaultFileDirectoryPath
	"Perform globbing expansion of a command argument. If argument is not a
	pathname, use aDefaultFileDirectory as the context for globbing. Answer a
	(possibly empty) collection of all matching paths relative to the default
	directory path."

	"ShellSyntax new expandArgument: 'ho?ts' inDirectoryPath: '/etc'"
	"ShellSyntax new expandArgument: '/b?n/s?' inDirectoryPath: '/etc'"
	"ShellSyntax new expandArgument: 'fs*' inDirectoryPath: '/etc'"
	"ShellSyntax new expandArgument: 'NOSUCHFILE' inDirectoryPath: '/etc'"
	"ShellSyntax new expandArgument: '/us*/b?n/v?' inDirectoryPath: '/etc'"
	"ShellSyntax new expandArgument: 'b?n/v?' inDirectoryPath: '/usr'"
	"ShellSyntax new expandArgument: nil inDirectoryPath: '/etc'"

	| dirPath pathNames newArgs |
	aString ifNil: [^ Array new: 0].
	dirPath := (aDefaultFileDirectoryPath
				ifNil: [self workingDirectory])
					ifNil: [FileDirectory default pathName].
	pathNames := (self containsExpansionTokens: aString)
				ifTrue: [self expandedPathsFrom: aString beginningAt: dirPath]
				ifFalse: [OrderedCollection new].
	newArgs := pathNames collect:
				[:p | (aDefaultFileDirectoryPath notNil and: [p beginsWith: aDefaultFileDirectoryPath, FileDirectory slash])
					ifTrue: [(ReadStream on: p) next: aDefaultFileDirectoryPath size + FileDirectory slash size; upToEnd]
					ifFalse: [p]].
	newArgs isEmpty ifTrue: [newArgs add: aString].
	^ newArgs
! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 12/27/2002 12:35'!
expandArgumentsFrom: argArray inDirectoryPath: aDefaultFileDirectoryPath
	"Perform globbing expansion of command arguments. For each argument which is not a
	pathname, use aDefaultFileDirectory as the context for globbing."

	"ShellSyntax new
		expandArgumentsFrom: #('-help' '/b?n/s?' 'ho?ts' 'fs*' 'NOSUCHFILE')
		inDirectoryPath: '/etc'"
	"ShellSyntax new
		expandArgumentsFrom: #('this' 'is' 'a' 'test' 'Sq*log' '/us*/b?n/v?')
		inDirectoryPath: FileDirectory default pathName"
	"ShellSyntax new
		expandArgumentsFrom: #('this' 'is' 'a' 'test' 'Sq*log' '/us*/b?n/v?')
		inDirectoryPath: nil"

	| newArgs |
	newArgs := OrderedCollection new.
	argArray do:
		[:arg |
		(self isFlagParameter: arg)
			ifTrue:
				[newArgs add: arg]
			ifFalse:
				[newArgs addAll: (self
					expandArgument: arg
					inDirectoryPath: aDefaultFileDirectoryPath)]].
	^ newArgs! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 12/17/2001 14:41'!
expandPathParts: pathParts beginningAt: root
	"Walk directory tree and answer a collection of matching path strings."

	| firstPart remainder dirPaths names dirNames basePath |
	(pathParts size == 0)
		ifTrue:
			["Should never get here, but just in case"
			^ Array with: root].
	firstPart := pathParts first.
	remainder := pathParts allButFirst.
	(firstPart = '.')
		ifTrue: "Stay at the current directory level"
			[dirPaths := Array with: root]
		ifFalse:
			[(firstPart = '..') "Move up a directory level if possible"
				ifTrue:
					[dirPaths := Array with: (self upOneDirectoryLevel: root)]
				ifFalse: "Normal case, a directory name"
					[remainder isEmpty
						ifTrue:
							["This condition terminates recursion"
							names := self glob: firstPart inDirectory: (FileDirectory on: root).
							^ (root = FileDirectory slash)
								ifTrue: [^ names collect: [:n | root, n]]
								ifFalse: [^ names collect: [:n | root, FileDirectory slash, n]]]
						ifFalse:
							["Continue recursion until no more elements"
							dirNames := self glob: firstPart
											names: (FileDirectory on: root) directoryNames.
							(('*', FileDirectory slash) match: root)
								ifTrue: [basePath := root]
								ifFalse: [basePath := root, FileDirectory slash].
							dirPaths := dirNames collect: [:d | basePath, d]]]].
		^ dirPaths
			inject: OrderedCollection new
			into: [:result :dir |
				result addAll: (self expandPathParts: remainder beginningAt: dir).
				result]
! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 12/28/2002 09:48'!
expandedPathsFrom: aPathString beginningAt: firstPartOfPath
	"Answer a collection of all paths which match aPathString. If aPathString is a relative path,
	expand the path relative to firstPartOfPath."

	 "ShellSyntax new expandedPathsFrom: '/e*c/h*s' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: 'e*c/h*s' beginningAt: '/'"
	 "ShellSyntax new expandedPathsFrom: '/usr/bin/vi' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: 'bin/vi' beginningAt: '/usr'"
	 "ShellSyntax new expandedPathsFrom: '/usr/*/vi' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: '/usr/bin/../bin/./../bin/../../usr/bin/vi' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: '/bin/../bin/./../bin/../../usr/bin/vi' beginningAt: '/usr'"
	 "ShellSyntax new expandedPathsFrom: '/usr/bin/../bin/./../bin/../../../../../usr/bin/vi' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: '../../../../../../usr/bin/vi' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: '/' beginningAt: '/'"
	 "ShellSyntax new expandedPathsFrom: '*.image' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: 'C:\WINDOWS\*.*' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: '*.*' beginningAt: 'C:\WINDOWS'"
	 "ShellSyntax new expandedPathsFrom: 'C:\*.*' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: 'C:\CONFIG.*' beginningAt: nil"
	 "ShellSyntax new expandedPathsFrom: 'C:\CONFIG.*' beginningAt: '\'"

	^ (self isAbsolutePath: aPathString)
		ifTrue: [self expandedPathsFromAbsolutePath: aPathString]
		ifFalse: [self expandedPathsFromRelativePath: aPathString beginningAt: firstPartOfPath]
! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 12/27/2002 17:26'!
expandedPathsFromAbsolutePath: anAbsolutePathString
	"Answer a collection of all paths which match anAbsolutePathString."

	 "ShellSyntax new expandedPathsFromAbsolutePath: '/e*c/h*s'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: 'e*c/h*s'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '/usr/bin/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: 'bin/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '/usr/*/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '/usr/bin/../bin/./../bin/../../usr/bin/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '/bin/../bin/./../bin/../../usr/bin/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '/usr/bin/../bin/./../bin/../../../../../usr/bin/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '../../../../../../usr/bin/vi'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '/'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '*.image'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: 'C:\WINDOWS\*.*'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: '*.*'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: 'C:\*.*'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: 'C:\CONFIG.*'"
	 "ShellSyntax new expandedPathsFromAbsolutePath: 'C:\CONFIG.*'"

	| result pathParts a prefix dirPathString root |
	result := OrderedCollection new.
	a := self splitLocationPrefixFromPath: anAbsolutePathString.
	prefix := a at: 1.
	dirPathString := a at: 2.
	pathParts := dirPathString findTokens: FileDirectory slash.
	root := prefix, FileDirectory slash.
	result addAll: (self expandPathParts: pathParts beginningAt: root).
	^ result

! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 12/27/2002 18:25'!
expandedPathsFromRelativePath: aPathString beginningAt: firstPartOfPath
	"Answer a collection of all paths which match aPathString. Expand the path relative to firstPartOfPath."

	 "ShellSyntax new expandedPathsFromRelativePath: 'e*c/h*s' beginningAt: '/'"
	 "ShellSyntax new expandedPathsFromRelativePath: 'usr/b?n/vi' beginningAt: '/'"
	 "ShellSyntax new expandedPathsFromRelativePath: 'bin/vi' beginningAt: '/usr'"
	 "ShellSyntax new expandedPathsFromRelativePath: 'usr/*/vi' beginningAt: nil"
	 "ShellSyntax new expandedPathsFromRelativePath: 'usr/bin/../bin/./../bin/../../usr/bin/vi' beginningAt: '/'"
	 "ShellSyntax new expandedPathsFromRelativePath: 'bin/../bin/./../bin/../../usr/bin/vi' beginningAt: '/usr'"
	 "ShellSyntax new expandedPathsFromRelativePath: '../../../../../../usr/bin/vi' beginningAt: '/'"
	 "ShellSyntax new expandedPathsFromRelativePath: '*.image' beginningAt: nil"
	 "ShellSyntax new expandedPathsFromRelativePath: '*.*' beginningAt: 'C:\WINDOWS'"

	| result pathParts root |
	result := OrderedCollection new.
	pathParts := aPathString findTokens: FileDirectory slash.
	root := firstPartOfPath isNil
				ifTrue: [FileDirectory default pathName]
				ifFalse: [firstPartOfPath].
	result addAll: (self expandPathParts: pathParts beginningAt: root).
	^ result

! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 10/13/2008 00:08'!
findExecutablePathFor: aString inDirectoryPath: pathString
	"Look for the executable, following Unix conventions for searching the PATH.
	If no likely candidate is found, answer nil."

	"ShellSyntax new findExecutablePathFor: 'bash' inDirectoryPath: nil"
	"ShellSyntax new findExecutablePathFor: 'ReadMe.txt' inDirectoryPath: nil"
	"ShellSyntax new findExecutablePathFor: 'SqueakDebug.log' inDirectoryPath: nil"
	"ShellSyntax new findExecutablePathFor: 'noSuchFileName' inDirectoryPath: nil"
	"ShellSyntax new findExecutablePathFor: 'a' inDirectoryPath: nil"
	"ShellSyntax new findExecutablePathFor: 'sqcat' inDirectoryPath: '/home/lewis/bin'"

	| delimiter path localNames fd fullName fileExists |
	self canAccessOSProcess ifFalse: [^ nil].
	delimiter := FileDirectory pathNameDelimiter.
	path := (pathString
				ifNil: [self workingDirectory])
					ifNil: [FileDirectory default pathName].
	fd := FileDirectory on: path.
	"Unix compatibility note: If aString contains wildcard characters,
	it is expanded in the context of the current working directory. Expansion
	does not take place in the context of the PATH directories."
	(aString includes: delimiter)
		ifTrue:
			["A fully or partially qualified path, do not search PATH"
			^ (self expandedPathsFrom: aString beginningAt: pathString)
				detect: [:fileName |
					(fd fileExists: fileName) and: [self isExecutable: fileName]]
				ifNone: [nil]]
		ifFalse:
			["Look for the file in the PATH locations"
			localNames := self glob: aString inDirectory: path.
			(localNames size == 0)
				ifTrue:
					["Name was not expanded, so use the unmodified value of aString"
					localNames := Array with: aString].
			(self pathEntriesInDirectoryPath: path)
				do: [:pathName | | thisDirectory |
					thisDirectory := FileDirectory on: pathName.
					localNames
						do: [:fileName |
							fileExists := [thisDirectory fileExists: fileName]
											on: InvalidDirectoryError
											do: [:e | false].
							fileExists ifTrue:
								[fullName := pathName, FileDirectory slash, fileName.
								(self isExecutable: fullName)
									ifTrue: [^ fullName]]]].
			^ nil]! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 9/12/2002 09:06'!
firstExpandedArgument: aString inDirectoryPath: aPath
	"Expand a single argument and answer a fully qualified path to the first matching
	name. If no match, answer the literal path and file name (which does not exist).
	If aString is nil, answer an empty string. This method may be used to expand a single
	argument, such as the target for a command output redirection, where the target is
	expected to resolve to a single file name."

	"ShellSyntax new firstExpandedArgument: 'b?*h' inDirectoryPath: '/usr/bin'"
	"ShellSyntax new firstExpandedArgument: 'b?*h' inDirectoryPath: '/no/such/directory'"
	"ShellSyntax new firstExpandedArgument: 'no*such*file???' inDirectoryPath: '/usr/bin'"
	"ShellSyntax new firstExpandedArgument: '/usr/bin/vi' inDirectoryPath: '/usr/bin'"
	"ShellSyntax new firstExpandedArgument: '../bin/vi' inDirectoryPath: '/usr/bin'"
	"ShellSyntax new firstExpandedArgument: '../b?n' inDirectoryPath: '/usr/bin'"
	"ShellSyntax new firstExpandedArgument: nil inDirectoryPath: '/usr/bin'"

	| expanded path trimmedString |
	aString ifNil: [^ ''].
	trimmedString := aString withBlanksTrimmed.
	expanded := self expandArgument: trimmedString inDirectoryPath: aPath.
	path := expanded isEmpty
			ifTrue: [trimmedString]
			ifFalse: [expanded first].
	^ self newPathFor: path fromDefault: aPath
! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 12/17/2001 09:52'!
getCwdForPath: aPathString

	| prefix |
	prefix := (self splitLocationPrefixFromPath: aPathString) first.
	^ prefix isEmpty
		ifTrue: [self workingDirectory]
		ifFalse: [self cwdDictionary at: prefix ifAbsentPut: [self defaultPathFor: aPathString]]
! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 10/13/2008 20:02'!
glob: aString
	"Apply Unix shell globbing rules to aString, and answer an array of matching
	file names."

	"ShellSyntax new glob: 'ReadMe.txt'"
	"ShellSyntax new glob: 'rEADmE.TXT'"
	"ShellSyntax new glob: 'R*dM?.txt'"

	^ self glob: aString inDirectory: self workingDirectory
! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 7/15/2002 08:16'!
glob: aString inDirectory: aFileDirectory
	"Apply Unix shell globbing rules to aString, and answer an array of matching
	file names."

	"ShellSyntax new glob: 'ReadMe.txt' inDirectory: FileDirectory default"
	"ShellSyntax new glob: 'SqueakDebug.log' inDirectory: FileDirectory default"
	"ShellSyntax new glob: '?queak*.[0-9].log' inDirectory: FileDirectory default"
	"ShellSyntax new glob: '*image*' inDirectory: FileDirectory default"
	"ShellSyntax new glob: 'noSuchFileName' inDirectory: FileDirectory default"
	"ShellSyntax new glob: 'b?n' inDirectory: (FileDirectory on: '/usr')"

	[^ self glob: aString names: aFileDirectory fileAndDirectoryNames]
		on: Error
		do: [:ex | ^ #()]

! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 3/11/2001 17:37'!
glob: aString names: fileNameArray
	"Apply Unix shell globbing rules to aString with respect to the elements of fileNameArray,
	and answer an array of matching file names. The fileNameArray is expected to have been
	provided by a FileDirectory query."

	| tokens |
	tokens := self globTokensFrom: aString.
	^ fileNameArray
		select: [:n |
			self tokenStream: (ReadStream on: tokens)
				matches: (ReadStream on: n)]

! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 3/10/2001 17:49'!
globRangeFrom: aString tokenStream: aTokenStream
	"Answer a string containing the characters specified by a range expression
	such as [a-z]. Sorry, this is just a simple implementation which assumes ASCII."

	| r w c t |
	(aString includes: $-) ifFalse: [^ aString]. "Simple case"
	r := ReadStream on: aString.
	w := WriteStream on: ''.
	(r peek == $!!) ifTrue:
		[r next.
		aTokenStream nextPut: #not].
	[r atEnd] whileFalse: [
		c := r next.
		(r peek == $-)
			ifTrue: [r next.
					t := r next.
					(c asciiValue to: t asciiValue) do:
						[:e | w nextPut: (Character value: e)]]
			ifFalse: [w nextPut: c]].
	^ w contents! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 12/17/2001 11:41'!
globTokensFrom: aString
	"Answer a collection of tokens to be used for glob matching."

	"ShellSyntax new globTokensFrom: 'abc?def\ghi\?jkl*mno\*pqr[hello]there[Xa-eY]Z[!!A-Z]end'"

	| lastToken r w t |
	lastToken := nil.
	r := ReadStream on: aString.
	w := WriteStream on: #().
	[r atEnd] whileFalse:
		[t := self nextGlobToken: r lastToken: lastToken tokenStream: w.
		(t ~~ #escape) ifTrue: [w nextPut: t].
		lastToken := t].
	^ w contents
! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 1/2/2002 19:54'!
isAbsolutePath: aPathString
	"Answer true if aPathString is specified relative to the root of the file system."

	"ShellSyntax new isAbsolutePath: 'not/absolute/path'"
	"ShellSyntax new isAbsolutePath: '/absolute/path'"
	"ShellSyntax new isAbsolutePath: '/'"
	"ShellSyntax new isAbsolutePath: 'C:\'"
	"ShellSyntax new isAbsolutePath: '\'"
	"ShellSyntax new isAbsolutePath: 'C:'"
	"ShellSyntax new isAbsolutePath: 'C:WINDOWS'"
	"ShellSyntax new isAbsolutePath: 'C:\WINDOWS'"

	| prefixAndPath path |
	^ ((((FileDirectory slash, '*') match: aPathString)
		or:	[prefixAndPath := self splitLocationPrefixFromPath: aPathString.
			(FileDirectory slash, '*') match: (path := prefixAndPath last)])
				or: [prefixAndPath first isEmpty not and: [path isEmpty]]) "One above the root of a Win32 file system tree"
! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 12/17/2001 11:40'!
isAtFileSystemRoot: aPathString
	"Answer true if aPathString is not a subdirectory."

	"ShellSyntax new isAtFileSystemRoot: 'D:\'"
	"ShellSyntax new isAtFileSystemRoot: 'D:'"
	"ShellSyntax new isAtFileSystemRoot: 'D:\Windows'"
	"ShellSyntax new isAtFileSystemRoot: '\'"

	^ (FileDirectory slash = (self splitLocationPrefixFromPath: aPathString) last) or: [aPathString isEmpty]
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 1/21/2007 14:07'!
isComment: aString
	"Answer true if aString is empty or if its first non-space character is commentChar"

	^ (aString size == 0)
		or:
			[aString do: [:e | (e isSeparator not)
							ifTrue: [^ e == self commentChar]].
			false]! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 12/15/2002 17:55'!
isDirectoryPath: aPathString

	| fullPath localName entry |
	^ (('*', FileDirectory slash) match: aPathString) or:
		[fullPath := FileDirectory default fullNameFor: aPathString.
		localName := FileDirectory localNameFor: fullPath.
		entry := (FileDirectory on: (FileDirectory dirPathFor: fullPath)) entries
			detect: [:e | e name = localName] ifNone: [nil].
		entry notNil and: [entry isDirectory]]
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 1/21/2007 14:10'!
isDoItCommand: aCommandString
	"Answer true if aCommandString should be evaluated as Smalltalk doIt."

	^ aCommandString withoutTrailingBlanks last == $!!! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 8/7/2002 13:36'!
isExecutable: aPath 
	"Answer true if aPath points to an executable file. This could be 
	enhanced to support setting UID and GID for a CommandShell session. 
	For now, just take the default for the current Squeak process."

	^ aPath notNil
		and: [(Smalltalk
				at: #OSProcess
				ifPresent: [:osp | osp accessor isExecutable: aPath])
				ifNil: [false]]! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 1/21/2007 14:12'!
isFlagParameter: aString
	"Answer true if aString begins with $- (the convention for command flags for Unix)"

	^ (aString size > 0) and: [(aString at: 1) == $-]! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 4/20/2003 11:31'!
newDirPath: aPath from: initialPath
	"Answer a path string for a change directory command. Change from initialPath
	(usually the current working directory) to the new path, where aPath may be
	either an absolute path name (fully specified from the root) or a relative path
	name (relative to initialPath)."

	"self new newDirPath: 'b?n' from: '/usr'"
	"self new newDirPath: '..' from: '/usr/bin'"
	"self new newDirPath: '.././bin' from: '/usr/bin'"
	"self new newDirPath: '../../../../..' from: '/usr/bin'"
	"self new newDirPath: '../../../../../usr' from: '/usr/bin'"

	^ (self expandedPathsFrom: aPath beginningAt: initialPath)
		at: 1
		ifAbsent: [nil]

! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 1/2/2002 22:09'!
newPathFor: aPathString fromDefault: startingPathString
	"Answer a new path starting from startingPathString if aPathString is a,
	relative path, otherwise just answer aPathString unmodified."

	"ShellSyntax new newPathFor: 'hosts' fromDefault: '/etc'"
	"ShellSyntax new newPathFor: 'three' fromDefault: 'one/two'"
	"ShellSyntax new newPathFor: 'one/two/three' fromDefault: ''"
	"ShellSyntax new newPathFor: '/hosts' fromDefault: '/etc'"
	"ShellSyntax new newPathFor: 'WINDOWS' fromDefault: 'C:\'"
	"ShellSyntax new newPathFor: 'D:\WINDOWS' fromDefault: 'C:\'"
	"ShellSyntax new newPathFor: 'D:WINDOWS' fromDefault: 'C:\'"

	^ (self isAbsolutePath: aPathString)
		ifTrue:
			[aPathString]
		ifFalse:
			[self appendPath: (self splitLocationPrefixFromPath: aPathString) last
					toPath: startingPathString]
! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 12/15/2001 09:55'!
nextGlobToken: aStream lastToken: aToken tokenStream: aTokenStream
	"Escape character processing probably belongs elsewhere. Since it can conflict
	with Win32 file name conventions, I have commented it out here. The code
	is otherwise left alone here in case I change my mind later. -dtl"

	| c |
	(aToken == #escape)
		ifTrue:
			[^ aStream next]
		ifFalse:
			[c := aStream next.
			(c == $[)
				ifTrue:
					["character range"
					^ self globRangeFrom: (aStream upTo: $]) tokenStream: aTokenStream]
				ifFalse:
					[(c == $*) ifTrue: [^ #matchRun].
					(c == $?) ifTrue: [^ #matchAny].
					"(c == $\) ifTrue: [^ #escape].	Escape processing disabled here"
					^ c asString]]! !

!ShellSyntax methodsFor: 'defaults' stamp: 'dtl 1/20/2007 14:11'!
noWaitToken
	"This character at the end of a command string means execute as detached process."

	^ '&'! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 4/20/2003 11:50'!
nullDirectoryString
	"String to display when the current directory is one level above the file system root."

	^ '[]'! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 10/30/2007 18:59'!
pathAboveRoot
	"Answer a path pointing to a location above the file system root. The
	pathStringAboveRoot string refers to this location."

	 "ShellSyntax new pathAboveRoot"

	^ (self platformName = 'Win32')
		ifTrue: ['']
		ifFalse: [FileDirectory slash].
! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 8/7/2002 16:13'!
pathEntriesInDirectoryPath: basePath
	"Answer a possibly empty collection of path entries from a colon-delimited path
	string provided by the operating system. Relative path names are interpreted relative
	to basePath"

	"ShellSyntax new pathEntriesInDirectoryPath: FileDirectory default pathName"

	^ (Smalltalk at: #OSProcess ifPresent:
		[:osp |
		(osp thisOSProcess path findTokens: ':')
			collect:
				[:e | self newPathFor: e fromDefault: basePath]])
		ifNil: [Array new].
! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 1/3/2002 19:17'!
pathRoot: aPathString
	"For aPathString pointing to a file in a file system, answer the root of the
	file system. This method works fine for Unix and Windows, but may require
	conditional code for other platforms."

	| prefix |
	prefix := (self splitLocationPrefixFromPath: aPathString) first.
	^ prefix, FileDirectory slash
! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 1/4/2002 21:19'!
pathStringAboveRoot
	"Magic string indicating location above the file system root. For Unix, this just
	points to the file system root. For Windows, it points to one level above the C: drive."

	^ '[]'! !

!ShellSyntax methodsFor: 'defaults' stamp: 'dtl 9/8/2001 10:16'!
pipeCharToken
	"The command pipeline character"

	^ '|'! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 10/30/2007 18:58'!
platformName 
	"After Squeak version 3.6, #platformName was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk platformName]) current platformName! !

!ShellSyntax methodsFor: 'path name expansion' stamp: 'dtl 10/13/2008 06:26'!
programNameAndArgumentsFrom: aString inDirectoryPath: pathString findExecutable: expandExePath
	"Answer an array with the name of the program to run, and with the expanded
	arguments list. If expandExePath is true, search the file system for an executable
	program to run, otherwise use the literal command name."

	"ShellSyntax new
		programNameAndArgumentsFrom: 'sleep 6'
		inDirectoryPath: '/etc'
		findExecutable: true"
	"ShellSyntax new
		programNameAndArgumentsFrom: 'sleep 6'
		inDirectoryPath: '/etc'
		findExecutable: false"
	"ShellSyntax new
		programNameAndArgumentsFrom: 'ls -l h?s*'
		inDirectoryPath: '/etc'
		findExecutable: true"

	| tokens tok args exe |
	tokens := self argsFromCommand: aString.
	tokens size == 0
		ifTrue: [tok := '']
		ifFalse: [tok := tokens at: 1].
	args := self
		expandArgumentsFrom: (tokens copyFrom: 2 to: tokens size)
		inDirectoryPath: pathString.
	exe := expandExePath
		ifTrue: [self findExecutablePathFor: tok inDirectoryPath: pathString]
		ifFalse: [tok].
	^ exe ifNotNil: [Array with: exe with: args]
! !

!ShellSyntax methodsFor: 'defaults' stamp: 'dtl 4/27/2002 18:57'!
quoteChar
	"In a Unix shell, this character is used to delimit chunks of text which should not
	be expanded with wildcards or shell variables. In a Smalltalk compiler, this character
	delimits a String. ShellSyntax interprets it as a delimiter for a  literal string which
	should not be expanded or interpreted in any way."

	^ $'! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 12/22/2007 10:17'!
redirectIOFor: commandString inDirectoryPath: dirPathString
	"Handle IO redirection. Answer a four element array with the command string
	(with IO tokens removed); and stdin, stdout and stderr if specified for
	redirection."

	"ShellSyntax new
		redirectIOFor: 'ls /tmp /NOSUCHFILE > /tmp/sqTempOut.tmp 2> /tmp/sqTempErr.tmp'
		inDirectoryPath: nil"
	"ShellSyntax new
		redirectIOFor: 'cat < /etc/hosts > /tmp/sqHosts.tmp'
		inDirectoryPath: nil"

	| cmd tokenStream tok target input output error |
	cmd := nil.
	input := output := error := nil.
	"Default to nil unless overridden by commandString."
	tokenStream := ReadStream on: (self tokenizeForRedirect: commandString).
	[tokenStream atEnd]
		whileFalse:
			[tok := tokenStream next.
			tok == #inputRedirect
				ifTrue: ["Do input redirection"
					target := self firstExpandedArgument: tokenStream next
								inDirectoryPath: dirPathString.
					"Use StandardFileStream to prevent multi-byte interpretation."
					input := StandardFileStream readOnlyFileNamed: target.
					input
						ifNil: [self inform: 'Cannot open ' , target , ', using default stdin']]
				ifFalse: [tok == #outputRedirect
						ifTrue: ["Do output redirection"
							target := self firstExpandedArgument: tokenStream next
										inDirectoryPath: dirPathString.
							output := self truncatedOutputFile: target.
							output ifNil:
									[self inform: 'Cannot open ' , target , ', using default stdout']]
						ifFalse: [tok == #outputAppendRedirect
								ifTrue: ["Do output redirection, appended to existing output file if any."
									target := self firstExpandedArgument: tokenStream next
												inDirectoryPath: dirPathString.
									output := FileStream fileNamed: target.
									output isNil
										ifTrue: [self inform: 'Cannot open ' , target , ', using default stdout']
										ifFalse: [output setToEnd]]
								ifFalse: [tok == #errorRedirect
										ifTrue: ["Do error redirection"
											target := self firstExpandedArgument: tokenStream next
														inDirectoryPath: dirPathString.
											error := self truncatedOutputFile: target.
											error ifNil:
													[self inform: 'Cannot open ' , target , ', using default stdout']]
										ifFalse: [tok == #errorAppendRedirect
												ifTrue: ["Do error redirection, appended to existing error file if any. "
													target := self firstExpandedArgument: tokenStream next
																inDirectoryPath: dirPathString.
													error := FileStream fileNamed: target.
													error isNil
														ifTrue: [self inform: 'Cannot open ' , target , ', using default stdout']
														ifFalse: [error setToEnd]]
												ifFalse: [cmd isNil
														ifTrue: [cmd := tok]
														ifFalse: [self inform: 'Ack!! Pfft!! IO redirection parsing error.']]]]]]].
	^ Array
		with: cmd
		with: input
		with: output
		with: error! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 12/16/2001 21:26'!
setCwdIndexFrom: aPathString
	"Set the current working directory index to a string (Win32 device name) or to #default."

	| prefix |
	prefix := (self splitLocationPrefixFromPath: aPathString) first.
	prefix isEmpty
		ifTrue: [self cwdIndex: #default]
		ifFalse: [self cwdIndex: prefix]
! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 10/30/2007 18:59'!
splitLocationPrefixFromPath: aPathString
	"Split any leading prefix from a path string, such as the C: device name
	for Win32. Answer an Array with prefix and path."

	 "ShellSyntax new splitLocationPrefixFromPath: 'C:\u'"
	 "ShellSyntax new splitLocationPrefixFromPath: 'C:\'"
	 "ShellSyntax new splitLocationPrefixFromPath: 'C:'"
	 "ShellSyntax new splitLocationPrefixFromPath: 'C:u'"
	 "ShellSyntax new splitLocationPrefixFromPath: 'c:u'"
	 "ShellSyntax new splitLocationPrefixFromPath: '\u'"

	| prefix path |
	(self platformName = 'Win32')
		ifTrue:
			[((aPathString at: 2 ifAbsent: []) == $:)
				ifTrue:
					[prefix := (String new: 2).
					prefix at: 1 put: (aPathString at: 1) asUppercase.
					prefix at: 2 put: $:.
					path := aPathString copyAfter: $:]
				ifFalse:
					[prefix := ''.
					path := aPathString]]
		ifFalse:
			[prefix := ''.
			path := aPathString].
	^ Array with: prefix with: path
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 1/20/2007 14:08'!
splitPipelineCommands: aCommand
	"Answer a list of the individual commands in a command pipeline string,
	delimited by the pipe character."

	^ (aCommand findTokens: self pipeCharToken) collect: [:e | e withBlanksTrimmed]
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 1/21/2007 15:11'!
stringTokensFrom: aString
	"Answer a two element array with a collection of strings from a whitespace
	delimited string, and with an optional error message. Single quote characters
	protect substrings from expansion to tokens. Single quote characters may be
	escaped by the backslash character to prevent this interpretation."

	| result r w errorMessage quoteEnabled t |
	result := OrderedCollection new.
	r := ReadStream on: aString.
	w := WriteStream on: String new.
	errorMessage := nil.
	(quoteEnabled := r peek == self quoteChar) ifTrue: [r next].
	[r atEnd] whileFalse:
		[t := r next.
		(t == self quoteChar)
			ifTrue:
				[quoteEnabled := quoteEnabled not]
			ifFalse:
				[quoteEnabled
					ifTrue:
						[w nextPut: t]
					ifFalse:
						[(t == self escapeChar)
							ifTrue:
								["An escaped character is treated as a literal character."
								t := r next.
								t isNil
									ifTrue: ["FIXME: This would indicate an escaped end of line"]
									ifFalse: [w nextPut: t]]
							ifFalse:
								[t isSeparator
									ifTrue:
										["Whitespace delimits tokens."
										(w isEmpty) ifFalse: [result add: w contents].
										w := WriteStream on: String new.
										[r peek notNil and: [r peek isSeparator]] whileTrue: [r next].]
									ifFalse:
										[w nextPut: t]]]]].
	quoteEnabled ifTrue: [errorMessage := 'unbalanced quotes'].
	(w isEmpty) ifFalse: [result add: w contents].
	^ Array with: result with: errorMessage
! !

!ShellSyntax methodsFor: 'globbing' stamp: 'dtl 1/5/2002 09:52'!
tokenStream: aTokenStream matches: aCharacterStream 
	"Iterate over both streams. If each character in aCharacterStream 
	matches its corresponding token in tokenStream, answer true."

	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'a?c')) matches: (ReadStream on: 'abc')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'ab*cde')) matches: (ReadStream on: 'abcde')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'ab*cde')) matches: (ReadStream on: 'abXXXcde')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'ab*cdX')) matches: (ReadStream on: 'abXXXcde')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'abc*')) matches: (ReadStream on: 'abcde')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'abc*')) matches: (ReadStream on: 'abc')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'a?cde*jkl')) matches: (ReadStream on: 'abcdefghijkl')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'abc[A-C]xyz')) matches: (ReadStream on: 'abcBxyz')"
	"ShellSyntax new tokenStream: (ReadStream on: (ShellSyntax new globTokensFrom: 'abc[!!A-C]xyz')) matches: (ReadStream on: 'abcZxyz')"

	| c tok |
	[aCharacterStream atEnd]
		whileFalse: [aTokenStream atEnd
				ifTrue: [^ false].
			c := aCharacterStream peek.
			tok := aTokenStream next.
			tok == #not
				ifTrue: ["A negated character range follows"
					tok := aTokenStream next.
					(self character: c matches: tok)
						ifTrue: [^ false].
					aCharacterStream next]
				ifFalse: ["The normal case"
					tok == #matchRun
						ifTrue:
							["Skip characters until we resync the streams, or answer false"
							[self tokenStream: aTokenStream copy matches: aCharacterStream copy]
								whileFalse:
									[aCharacterStream atEnd ifTrue: [^ false].
									aCharacterStream next]]
						ifFalse: [(self character: c matches: tok)
								ifFalse: [^ false].
							aCharacterStream next]]].
	^ aTokenStream atEnd
		or: [aTokenStream next == #matchRun
				and: [aTokenStream atEnd]]
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 12/17/2001 11:39'!
tokenizeForRedirect: aString
	"Answer a collection of tokens for IO redirection. Answers a collection of
	tokens, where each token is either a String, or a Symbol in:
	#(#inputRedirect #outputRedirect #outputAppendRedirect #errorRedirect #errorAppendRedirect)"

	"ShellSyntax new tokenizeForRedirect: 'one > two < three 2> four five six 2>> seven eight > nine ten >> eleven twelve 2 > thirteen'"

	| oc rs ws c redirectToken tok |
	oc := OrderedCollection new.
	rs := ReadStream on: aString.
	ws := WriteStream on: ''.
	[rs atEnd]
		whileFalse:
			[c := rs next.
			((c == $2) and: [rs peek == $>])
				ifTrue:
					[rs next.
					(rs peek == $>)
						ifTrue:
							[rs next.
							redirectToken := #errorAppendRedirect]
						ifFalse:
							[redirectToken := #errorRedirect].
					c := ws contents.
					(c size > 0)
						ifTrue:
							[oc add: c.
							ws resetContents].
					oc add: redirectToken]
				ifFalse:
					[(c == $>)
						ifTrue:
							[(rs peek == $>)
								ifTrue:
									[rs next.
									redirectToken := #outputAppendRedirect]
								ifFalse:
									[redirectToken := #outputRedirect].
							c := ws contents.
							(c size > 0)
								ifTrue:
									[oc add: c.
									ws resetContents].
							oc add: redirectToken]
						ifFalse:
							[(c == $<)
								ifTrue:
									[c := ws contents.
									(c size > 0)
										ifTrue:
											[oc add: c.
											ws resetContents].
									oc add: #inputRedirect]
								ifFalse: [ws nextPut: c]]]].
	tok := ws contents.
	(tok size > 0) ifTrue: [oc add: tok].
	^ oc
! !

!ShellSyntax methodsFor: 'command string parsing' stamp: 'dtl 12/17/2001 11:39'!
truncatedOutputFile: fileName
	"Attempt to do reasonable things to open a file of size zero. Better would be
	to implement a file truncation primitive."

	"ShellSyntax new truncatedOutputFile: '/tmp/delete.me'"
	"ShellSyntax new truncatedOutputFile: '/dev/null'"

	| strm |
	[strm := FileStream concreteStream forceNewFileNamed: fileName]
		on: Error
		do: [:ex | ^ FileStream fileNamed: fileName].
	^ strm
! !

!ShellSyntax methodsFor: 'platform dependent' stamp: 'dtl 12/27/2002 11:06'!
upOneDirectoryLevel: aPathName
	"Answer a path name for a directory one level up the directory tree"

	 "ShellSyntax new upOneDirectoryLevel: 'C:\Windows\temp'"
	 "ShellSyntax new upOneDirectoryLevel: 'C:\Windows'"
	 "ShellSyntax new upOneDirectoryLevel: 'C:\'"
	 "ShellSyntax new upOneDirectoryLevel: ''"
	 "ShellSyntax new upOneDirectoryLevel: '/usr/bin'"
	 "ShellSyntax new upOneDirectoryLevel: '/usr'"
	 "ShellSyntax new upOneDirectoryLevel: '/'"
	 "ShellSyntax new upOneDirectoryLevel: 'usr'"
	 "ShellSyntax new upOneDirectoryLevel: ''"

	| prefixAndPath prefix path newPath |
	prefixAndPath := self splitLocationPrefixFromPath: aPathName.
	prefix := prefixAndPath at: 1.
	path := prefixAndPath at: 2.
	(self isAtFileSystemRoot: aPathName)
		ifTrue:
			[(self isAbsolutePath: aPathName)
				ifTrue:
					[prefix isEmpty
						ifTrue: "Normal case, unix. Do not go up a level."
							[^ aPathName]
						ifFalse: "A Win32 device name, go up a level above the file systems"
							[^ '']]
				ifFalse:
					["One level above whererever we are. Unfortunately we have
					no way to know where we are, so this probably should be treated
					as a sytax error."
					^ '..']]
		ifFalse:
			[newPath := (path includes: FileDirectory pathNameDelimiter)
				ifTrue: [path copyUpToLast: FileDirectory pathNameDelimiter]
				ifFalse: [''].
			^ newPath isEmpty
				ifTrue:
					[(self isAbsolutePath: aPathName)
						ifTrue: [prefix, FileDirectory slash]
						ifFalse: [prefix]]
				ifFalse:
					[prefix, newPath]]
! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 8/7/2002 13:32'!
workingDirectory

	^ self cwdDictionary at: self cwdIndex ifAbsentPut:
		[Smalltalk at: #OSProcess ifPresent: [:osp | osp thisOSProcess getCwd]]
! !

!ShellSyntax methodsFor: 'working directory' stamp: 'dtl 12/16/2001 21:32'!
workingDirectory: aPathString

	self setCwdIndexFrom: aPathString.
	self cwdDictionary at: self cwdIndex put: aPathString! !
CommandShell initialize!
